###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Per_Interaction_Dimensions_Schema = [{
    'name': 'Interaction_Number',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Date_Time',
    'type': 'DATETIME',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser_Group',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Asset',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Asset_Category',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Asset_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Asset_Orientation',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Audience_Targeted',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_End_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_Start_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Groups_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Groups_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Pixel_Size',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Version',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1_Value_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2_Value_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3_Value_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4_Value_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5_Value_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Profile',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Profile_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Attributed_Interaction',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Channel',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Engine_Country',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Engine_Property',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Engine_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Landing_Page',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Landing_Page_Query_String',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Processed_Landing_Page',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Processed_Landing_Page_Query_String',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Query',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Package_Roadblock',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Package_Roadblock_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Package_Roadblock_Strategy',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad_Group',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad_Group_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Advertiser',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Advertiser_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Agency',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Agency_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Bid_Strategy',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Bid_Strategy_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Campaign',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Campaign_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Engine_Account',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Engine_Account_Category',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Engine_Account_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_External_Ad_Group_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_External_Ad_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_External_Campaign_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_External_Keyword_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Keyword',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Keyword_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Labels',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Landing_Page_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Legacy_Keyword_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Match_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_End_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Pixel_Size',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Start_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Recalculated_Attributed_Interaction',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Rendering_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Backup_Image',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Counters',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Exits',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Timers',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Display_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Expansion_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Expansions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Video_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Video_Plays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Html5_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Interaction_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interactive_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Manual_Closes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Companion_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_First_Quartile_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Full_Screen',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Interactions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Length',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Midpoints',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Mutes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Pauses',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Plays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Progress_Events',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Replays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Skips',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Stops',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Third_Quartile_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Unmutes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Views',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Dcm',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Site_Directory',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Id_Site_Directory',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Id_Dcm',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Keyname',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Twitter_Impression_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Within_Floodlight_Lookback_Window',
    'type': 'STRING',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Module that centralizes all CM data access."""

from starthinker.util.google_api import API_DCM
from starthinker.task.traffic.store import store


class BaseDAO(object):
  """Parent class to all data access objects.

  Centralizes all common logic to all classes that access CM to create or update
  entities.
  """

  def __init__(self, auth, profile_id, is_admin=False):
    """Initializes the object with a specific CM profile ID and an authorization scheme."""
    self.auth = auth
    self.is_admin = is_admin
    self.profile_id = profile_id
    self._entity = 'UNDEFINED'
    self._metrics = {}

  def _api(self, iterate=False):
    """Returns an DCM API instance.

    Must be overloaded by one of the derived classes and extended for
    specific endpoint.
    """
    return API_DCM(self.auth, iterate, self.is_admin)

  def _clean(self, item):
    """Removes null keys from the item.

    An absent key and a null key mean different things in certain contexts for
    CM, this method cleans up objects before sending to the CM API by removing
    any null keys.

    Args:
      item: The CM object to clean.
    """
    # Write code here to remove all null fields from item
    null_keys = []
    for key in item:
      if item[key] == None:
        null_keys.append(key)

    for key in null_keys:
      del item[key]

    return item

  def _get(self, feed_item):
    """Fetches an item from CM.

    Args:
      feed_item: Feed item from Bulkdozer feed representing the item to fetch
        from CM.
    """
    #print('hitting the api to get %s, %s' % (self._entity, feed_item[self._id_field]))
    return self._api().get(
        profileId=self.profile_id, id=feed_item[self._id_field]).execute()

  def get(self, feed_item, required=False, column_name=None):
    """Retrieves an item.

    Items could be retrieved from a in memory cache in case it has already been
    retrieved within the current execution. Also, this method is capable of
    translating 'ext' placeholder IDs with concrete CM ids.

    Args:
      feed_item: Feed item from the Bulkdozer feed representing the item to
        retrieve.

    Returns:
      The CM object that represents the identified entity.
    """
    result = None
    keys = []
    id_value = feed_item.get(self._id_field,
                             None) if column_name == None else feed_item.get(
                                 column_name, None)

    if not id_value and self._search_field and feed_item.get(
        self._search_field, None):
      store_key = feed_item[self._search_field]

      if self._parent_filter_name:
        if feed_item.get(self._parent_filter_field_name, None):
          store_key = str(feed_item.get(self._parent_filter_field_name,
                                        None)) + store_key

      result = store.get(self._entity, store_key)

      if not result:
        result, key = self._get_by_name(feed_item)
        keys.append(key)

      if not result and required:
        raise Exception('ERROR: Could not find %s with name %s' %
                        (self._entity, feed_item[self._search_field]))
    elif id_value:
      if isinstance(id_value, str) and id_value.startswith('ext'):
        keys.append(id_value)
        id_value = store.translate(self._entity, id_value)

        if id_value and not column_name:
          feed_item[self._id_field] = id_value

        elif id_value and column_name:
          feed_item[column_name] = id_value

      if id_value:
        keys.append(id_value)
        result = store.get(self._entity, id_value)

        if not result:
          result = self._get(feed_item)

        if not result and required:
          raise Exception('ERROR: Could not find %s with id %s' %
                          (self._entity, id_value))

    store.set(self._entity, keys, result)

    return result

  def _get_base_search_args(self, search_string):
    return {
        'profileId': self.profile_id,
        'searchString': search_string,
        'sortField': 'NAME',
        'maxResults': 2
    }

  def _get_by_name(self, feed_item):
    """Searches CM for an item of name defined in the search field of the DAO class.

    If more than one item is returned an error is raised, e.g. if there are more
    than one item with the same name.

    Args:
      feed_item: The Bulkdozer feed item with the name to search for.

    Returns:
      If found, the CM entity object that matches the search string.
    """
    key = ''
    if self._search_field:
      key = feed_item[self._search_field].strip()

      search_string = feed_item[self._search_field].strip()
      args = self._get_base_search_args(search_string)

      if self._parent_filter_name:
        if feed_item.get(self._parent_filter_field_name, None):
          args[self._parent_filter_name] = feed_item.get(
              self._parent_filter_field_name, None)
        elif self._parent_dao:
          parent = self._parent_dao.get(feed_item, required=True)
          if parent:
            args[self._parent_filter_name] = parent.get('id', None)

        key = str(args.get(self._parent_filter_name, '')) + key

      print('hitting the api to search for %s, %s' %
            (self._entity, search_string))
      search_result = self._api().list(**args).execute()

      items = search_result[self._list_name]

      if items and len(items) > 0:
        item = items[0]

        if search_string == item['name']:
          if len(items) > 1 and items[1]['name'] == search_string:
            raise Exception('ERROR: More than one item found with %s %s' %
                            (self._search_field, feed_item[self._search_field]))
          else:
            return item, key

    return None, key

  def _insert(self, item, feed_item):
    """Inserts a new item into CM.

    Args:
      item: The CM object to insert.
      feed_item: The feed item from the Bulkdozer feed representing the item to
        insert.

    Returns:
      The CM object representing the item inserted.
    """
    #print('ITEM', item)
    return self._api().insert(profileId=self.profile_id, body=item).execute()

  def _update(self, item, feed_item):
    """Updates a new item in CM.

    Args:
      item: The CM object to update.
      feed_item: The feed item from the Bulkdozer feed representing the item to
        update.
    """
    #print('ITEM', item)
    self._api().update(profileId=self.profile_id, body=item).execute()

  def pre_fetch(self, feed):
    """Pre-fetches all required items to be update into the cache.

    This increases performance for update operations.

    Args:
      feed: List of feed items to retrieve
    """
    if hasattr(self, '_list_name') and self._list_name and self._id_field:
      print('pre fetching %s' % self._list_name)
      ids = [
          feed_item[self._id_field]
          for feed_item in feed
          if isinstance(feed_item[self._id_field], int)
      ]

      if ids:
        for i in range(0, len(ids), 500):
          results = self._api(iterate=True).list(
              profileId=self.profile_id, ids=ids[i:i + 500]).execute()
          for item in results:
            store.set(self._entity, [item['id']], item)

  def process(self, feed_item):
    """Processes a Bulkdozer feed item.

    This method identifies if the item needs to be inserted or updated, cleans
    it, performs the CM operations required, and update the feed item with newly
    created ids and name lookups so that the feed can be updated.

    Args:
      feed_item: Bulkdozer feed item to process.

    Returns:
      Newly created or updated CM object.
    """
    item = self.get(feed_item)

    if item:
      self._process_update(item, feed_item)

      self._clean(item)

      self._update(item, feed_item)
    else:
      new_item = self._process_new(feed_item)

      self._clean(new_item)

      item = self._insert(new_item, feed_item)

      if self._id_field and feed_item.get(self._id_field, '').startswith('ext'):
        store.map(self._entity, feed_item.get(self._id_field), item['id'])
        store.set(self._entity, [feed_item[self._id_field]], item)

      if self._search_field and feed_item.get(self._search_field, ''):
        store.map(self._entity, feed_item.get(self._search_field), item['id'])
        store.set(self._entity, [feed_item[self._search_field]], item)

    if item:
      feed_item[self._id_field] = item['id']
      store.set(self._entity, [item['id']], item)

    self._post_process(feed_item, item)

    return item

  def _post_process(self, feed_item, item):
    """Provides an opportunity for sub classes to perform any required operations after the item has been processed.

    Args:
      feed_item: The Bulkdozer feed item that was processed.
      item: The CM object resulting from the process operation.
    """
    pass



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from starthinker_ui.recipe.models import Recipe
from django.core import serializers


class Command(BaseCommand):
  help = 'Shows status of recipes'

  def add_arguments(self, parser):
    parser.add_argument(
        '--recipe',
        action='store',
        dest='recipe',
        default=None,
        help='Run a specific recipe.',
    )

    parser.add_argument(
        '--raw',
        action='store_true',
        dest='raw',
        default=False,
        help='Raw recipe log.',
    )

  def handle(self, *args, **kwargs):
    for recipe in (Recipe.objects.filter(
        pk=kwargs['recipe']) if kwargs['recipe'] else Recipe.objects.all()):
      status = json.loads(recipe.job_status)

      print('---------------------------------------')
      print('Name:', recipe.name)
      print('Account:', recipe.account.email)
      print('UUID:', recipe.uid())
      print('Reference:', recipe.reference)
      print('Active:', recipe.active)
      print('Week:', recipe.week)
      print('Hour:', recipe.hour)
      print('Timezone:', recipe.timezone)
      print('Manual:', recipe.manual)
      print('Done:', all(task['done'] for task in status['tasks']))

      if kwargs['raw']:
        print('Job Status:')
        print(json.dumps(status), indent=2)
        continue

      log = recipe.get_log()

      print('Tasks', len(log['tasks']))
      print('UTC', log['utc'])
      print('AGO', log['ago'])
      print('PK:', recipe.pk)
      print('Date Timezone', log['date_tz'])
      print('Forced', log.get('forced', False))
      print('Status', log['status'])
      print('Worker:', recipe.worker_uid)

      for task in log['tasks']:
        print('  ----------------')
        print('  Utc', task['utc'])
        print('  Ago', task['ago'])
        print('  Script', task['script'])
        print('  Instance', task['instance'])
        print('  Hour', task['hour'])
        print('  Event', task['event'])
        print('  Done', task['done'])
        print('  Output', task['stdout'])
        print('  Error', task['stderr'])
        print('')

      task = recipe.get_task()
      print('Next Task', task)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

ABERRANT_PLURAL_MAP = {
    'appendix': 'appendices',
    'barracks': 'barracks',
    'cactus': 'cacti',
    'child': 'children',
    'criterion': 'criteria',
    'deer': 'deer',
    'echo': 'echoes',
    'elf': 'elves',
    'embargo': 'embargoes',
    'focus': 'foci',
    'fungus': 'fungi',
    'goose': 'geese',
    'hero': 'heroes',
    'hoof': 'hooves',
    'index': 'indices',
    'knife': 'knives',
    'leaf': 'leaves',
    'life': 'lives',
    'man': 'men',
    'mouse': 'mice',
    'nucleus': 'nuclei',
    'person': 'people',
    'phenomenon': 'phenomena',
    'potato': 'potatoes',
    'self': 'selves',
    'syllabus': 'syllabi',
    'tomato': 'tomatoes',
    'torpedo': 'torpedoes',
    'veto': 'vetoes',
    'woman': 'women',
}

VOWELS = set('aeiou')


def pluralize(singular):
  """Return plural form of given lowercase singular word (English only).

  Based on
  ActiveState recipe http://code.activestate.com/recipes/413172/

  >>> pluralize('')
  ''
  >>> pluralize('goose')
  'geese'
  >>> pluralize('dolly')
  'dollies'
  >>> pluralize('genius')
  'genii'
  >>> pluralize('jones')
  'joneses'
  >>> pluralize('pass')
  'passes'
  >>> pluralize('zero')
  'zeros'
  >>> pluralize('casino')
  'casinos'
  >>> pluralize('hero')
  'heroes'
  >>> pluralize('church')
  'churches'
  >>> pluralize('x')
  'xs'
  >>> pluralize('car')
  'cars'

  """
  if not singular:
    return ''
  plural = ABERRANT_PLURAL_MAP.get(singular)
  if plural:
    return plural
  root = singular
  try:
    if singular[-1] == 'y' and singular[-2] not in VOWELS:
      root = singular[:-1]
      suffix = 'ies'
    elif singular[-1] == 's':
      if singular[-2] in VOWELS:
        if singular[-3:] == 'ius':
          root = singular[:-2]
          suffix = 'i'
        else:
          root = singular[:-1]
          suffix = 'ses'
      else:
        suffix = 'es'
    elif singular[-2:] in ('ch', 'sh'):
      suffix = 'es'
    else:
      suffix = 's'
  except IndexError:
    suffix = 's'
  plural = root + suffix
  return plural


if __name__ == '__main__':
  import doctest
  doctest.testmod()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of Creatives."""

import json

from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.feed import FieldMap
from starthinker.task.traffic.creative_assets import CreativeAssetDAO
from starthinker.task.traffic.landing_page import LandingPageDAO


class CreativeDAO(BaseDAO):
  """Creative data access object.

  Inherits from BaseDAO and implements creative specific logic for creating and
  updating creatives.
  """

  def __init__(self, auth, profile_id, is_admin):
    """Initializes CreativeDAO with profile id and authentication scheme."""
    super(CreativeDAO, self).__init__(auth, profile_id, is_admin)

    self._entity = 'CREATIVE'

    self._parent_filter_name = 'advertiserId'
    self._parent_filter_field_name = FieldMap.ADVERTISER_ID

    self._id_field = FieldMap.CREATIVE_ID
    self._search_field = FieldMap.CREATIVE_NAME
    self._list_name = 'creatives'
    self._parent_dao = None

    self.creative_asset_dao = CreativeAssetDAO(auth, profile_id, is_admin, None)
    self.landing_page_dao = LandingPageDAO(auth, profile_id, is_admin)

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(CreativeDAO, self)._api(iterate).creatives()

  def _assignment_matches(self, item, assignment):
    if item.get(FieldMap.CREATIVE_ID, None) and assignment.get(
        FieldMap.CREATIVE_ID, None):
      return item.get(FieldMap.CREATIVE_ID,
                      None) == assignment.get(FieldMap.CREATIVE_ID, None)
    else:
      return item.get(FieldMap.CREATIVE_NAME,
                      '1') == assignment.get(FieldMap.CREATIVE_NAME, '2')

  def map_creative_third_party_url_feeds(self, creative_feed,
                                         third_party_url_feed):
    """Maps third party url feed to the corresponding creative.

    Third party URL is a child object to the creative, and there is a 1 creative
    to many third party urls relationship. In Bulkdozer they are represented by
    two separate tab in the feed, and this method maps the creatives to their
    respective third party URLs based on the creative ID.

    Args:
      creative_feed: Creative feed.
      third_party_url_feed: Third party url feed.
    """
    for creative in creative_feed:
      creative['third_party_urls'] = [
          third_party_url for third_party_url in third_party_url_feed
          if self._assignment_matches(creative, third_party_url)
      ]

  def map_creative_click_tag_feeds(self, creative_feed, click_tag_feed):
    """Maps click tag feed to the corresponding creative.

    Click Tag is a child object to the creative, and there is a 1 creative
    to many click tags relationship. In Bulkdozer they are represented by
    two separate tab in the feed, and this method maps the creatives to their
    respective click tags based on the creative ID.

    Args:
      creative_feed: Creative feed.
      click_tag_feed: Click tag feed.
    """
    for creative in creative_feed:
      creative['click_tags'] = [
          click_tag for click_tag in click_tag_feed
          if self._assignment_matches(creative, click_tag)
      ]

  def map_creative_and_association_feeds(self, creative_feed,
                                         creative_association_feed):
    """Maps creative association feed to the corresponding creative.

    Creative association is a child object to the creative, and there is a 1
    creative to many creative association relationship. In Bulkdozer they are
    represented by two separate tab in the feed, and this method maps the
    creatives to their respective creative association based on the creative ID.

    Args:
      creative_feed: Creative feed.
      creative_association_feed: Creative association feed.
    """
    for creative in creative_feed:
      creative['associations'] = [
          association for association in creative_association_feed
          if self._assignment_matches(creative, association)
      ]

  def _associate_third_party_urls(self, feed_item, creative):
    """Associate third party urls with the respective creative DCM object.

    This method transforms all child feed mapped earlier into DCM formatted
    associations within the creative object so it can be pushed to the API.

    Args:
      feed_item: Feed item representing the creative.
      creative: DCM creative object being created or updated.
    """
    third_party_urls = []
    for third_party_url in feed_item.get('third_party_urls', []):
      third_party_url_type = FieldMap.THIRD_PARTY_URL_TYPE_MAP.get(
          third_party_url.get(FieldMap.THIRD_PARTY_URL_TYPE, None))
      if third_party_url_type:
        third_party_urls.append({
            'thirdPartyUrlType': third_party_url_type,
            'url': third_party_url.get(FieldMap.THIRD_PARTY_URL, None)
        })

    if third_party_urls:
      creative['thirdPartyUrls'] = third_party_urls

  def _associate_click_tags(self, feed_item, creative):
    """Associate click tags with the respective creative DCM object.

    This method transforms all child feed mapped earlier into DCM formatted
    associations within the creative object so it can be pushed to the API.

    Args:
      feed_item: Feed item representing the creative.
      creative: DCM creative object being created or updated.
    """
    click_tags = []
    for click_tag in feed_item.get('click_tags', []):
      lp = self.landing_page_dao.get(
          click_tag, column_name=FieldMap.CLICK_TAG_LANDING_PAGE_ID)

      ct = {
          'eventName': click_tag.get(FieldMap.CLICK_TAG_EVENT, None),
          'name': click_tag.get(FieldMap.CLICK_TAG_NAME, None),
          'clickThroughUrl': {}
      }

      if (click_tag.get(FieldMap.CLICK_TAG_LANDING_PAGE_ID)):
        ct['clickThroughUrl']['landingPageId'] = click_tag.get(
            FieldMap.CLICK_TAG_LANDING_PAGE_ID) if not lp else lp['id']
      elif (click_tag.get(FieldMap.CLICK_TAG_CUSTOM_CLICK_THROUGH_URL)):
        ct['clickThroughUrl']['customClickThroughUrl'] = click_tag.get(
            FieldMap.CLICK_TAG_CUSTOM_CLICK_THROUGH_URL)

      click_tags.append(ct)

    if click_tags:
      creative['clickTags'] = click_tags

  def _process_update(self, item, feed_item):
    """Updates a creative based on the values from the feed.

    Args:
      item: Object representing the creative to be updated, this object is
        updated directly.
      feed_item: Feed item representing creative values from the Bulkdozer feed.
    """
    item['name'] = feed_item.get(FieldMap.CREATIVE_NAME, None)

    self._associate_third_party_urls(feed_item, item)

    self._associate_click_tags(feed_item, item)

  def _process_new(self, feed_item):
    """Creates a new creative DCM object from a feed item representing an creative from the Bulkdozer feed.

    This function simply creates the object to be inserted later by the BaseDAO
    object.

    Args:
      feed_item: Feed item representing the creative from the Bulkdozer feed.

    Returns:
      A creative object ready to be inserted in DCM through the API.

    """
    creative = {
        'advertiserId': feed_item.get(FieldMap.ADVERTISER_ID, None),
        'name': feed_item.get(FieldMap.CREATIVE_NAME, None),
        'active': True
    }

    self._associate_third_party_urls(feed_item, creative)
    self._associate_click_tags(feed_item, creative)

    # Video Creatives
    if feed_item.get(FieldMap.CREATIVE_TYPE, None) == 'VIDEO':
      creative['type'] = 'INSTREAM_VIDEO'

      for association in feed_item.get('associations', []):
        identifier = self.creative_asset_dao.get_identifier(
            association, self._creative_asset_feed)
        creative['creativeAssets'] = [{
            'assetIdentifier': identifier,
            'role': 'PARENT_VIDEO'
        }]

      del creative['active']

    #Display Creatives
    elif feed_item.get(FieldMap.CREATIVE_TYPE, None) == 'DISPLAY':
      creative['type'] = 'DISPLAY'

      # Creative Size
      if feed_item.get(FieldMap.CREATIVE_WIDTH, None) and feed_item.get(
          FieldMap.CREATIVE_HEIGHT, None):
        creative['size'] = {
            'kind': 'dfareporting#size',
            'width': feed_item.get(FieldMap.CREATIVE_WIDTH, None),
            'height': feed_item.get(FieldMap.CREATIVE_HEIGHT, None)
        }

      # Add Creative Assets
      for association in feed_item.get('associations', []):
        identifier = self.creative_asset_dao.get_identifier(
            association, self._creative_asset_feed)
        creative['creativeAssets'] = [{
            'assetIdentifier': identifier,
            'role': 'PRIMARY'
        }]

        # Backup Asset
        if feed_item.get(FieldMap.CREATIVE_BACKUP_ASSET_ID,
                         None) and feed_item.get(
                             FieldMap.CREATIVE_BACKUP_ASSET_ID, None) != '':
          backup_identifier = self.creative_asset_dao.get_backup_identifier(
              association, self._creative_asset_feed)

          creative['backupImageReportingLabel'] = feed_item.get(
              FieldMap.CREATIVE_BACKUP_NAME, None)

          # Parse Features
          backup_features = feed_item.get(FieldMap.BACKUP_IMAGE_FEATURES, None)
          if backup_features != None or backup_features != '':
            features = backup_features.split(',')
            creative['backupImageFeatures'] = features

          creative['backupImageTargetWindow'] = {
              'targetWindowOption':
                  feed_item.get(FieldMap.BACKUP_IMAGE_TARGET_WINDOW_OPTION,
                                None),
              'customHtml':
                  feed_item.get(FieldMap.BACKUP_IMAGE_CUSTOM_HTML, None)
          }

          lp = self.landing_page_dao.get(
              feed_item,
              column_name=FieldMap.BACKUP_IMAGE_CLICK_THROUGH_LANDING_PAGE_ID)
          creative['backupImageClickThroughUrl'] = {
              'landingPageId':
                  feed_item.get(
                      FieldMap.BACKUP_IMAGE_CLICK_THROUGH_LANDING_PAGE_ID)
                  if not lp else lp['id']
          }

          # Backup Image
          creative['creativeAssets'].append({
              'assetIdentifier': backup_identifier,
              'role': 'BACKUP_IMAGE'
          })

      del creative['active']

    else:
      raise Exception('Only video and display are supported at the moment!')
    # (mauriciod@): I didn't pull the display creative stuff from jeltz in here,
    # because I am splitting things up differently, and the backup image will
    # have to be uploaded in the creative_assets dao

    return creative

  def map_assets_feed(self, creative_asset_feed):
    self._creative_asset_feed = creative_asset_feed

  def _post_process(self, feed_item, new_item):
    """Maps ids and names of related entities so they can be updated in the Bulkdozer feed.

    When Bulkdozer is done processing an item, it writes back the updated names
    and ids of related objects, this method makes sure those are updated in the
    creative feed.

    Args:
      feed_item: Feed item representing the creative from the Bulkdozer feed.
      item: The DCM creative being updated or created.
    """
    # TODO loop through 3p urls and update the feed
    for third_party_url in feed_item.get('third_party_urls', []):
      third_party_url[FieldMap.CREATIVE_ID] = new_item['id']
      third_party_url[FieldMap.CREATIVE_NAME] = new_item['name']

    for association in feed_item.get('associations', []):
      association[FieldMap.CREATIVE_ID] = self.get(association)['id']
      association[FieldMap.CREATIVE_NAME] = self.get(association)['name']

      dcm_association = self.creative_asset_dao.get(association, required=True)
      if dcm_association:
        association[FieldMap.CREATIVE_ASSET_ID] = dcm_association.get(
            'id', None)
        association[FieldMap.CREATIVE_ASSET_NAME] = dcm_association.get(
            'name', None)
        backup_lp = self.landing_page_dao.get(
            feed_item,
            column_name=FieldMap.BACKUP_IMAGE_CLICK_THROUGH_LANDING_PAGE_ID)
        if backup_lp:
          association[
              FieldMap
              .BACKUP_IMAGE_CLICK_THROUGH_LANDING_PAGE_ID] = backup_lp['id']
          association[
              FieldMap
              .BACKUP_IMAGE_CLICK_THROUGH_LANDING_PAGE_NAME] = backup_lp['name']
        # Backup Asset Id
        backup_asset = self.creative_asset_dao.get(
            association, column_name=FieldMap.CREATIVE_BACKUP_ASSET_ID)

        if backup_asset:
          association[FieldMap.CREATIVE_BACKUP_ASSET_ID] = backup_asset['id']

        # association[FieldMap.BACKUP_ASSET_ID] = dcm_association.get(
        #     'id', None)

    # Update Click Tags
    for click_tag in feed_item.get('click_tags', []):
      #Update Creative info
      click_tag[FieldMap.CREATIVE_ID] = new_item['id']
      click_tag[FieldMap.CREATIVE_NAME] = new_item['name']

      # Update Landing Page
      click_tag_lp = self.landing_page_dao.get(
          click_tag, column_name=FieldMap.CLICK_TAG_LANDING_PAGE_ID)
      if click_tag_lp:
        click_tag[FieldMap.CLICK_TAG_LANDING_PAGE_ID] = click_tag_lp['id']
        click_tag[FieldMap.CLICK_TAG_LANDING_PAGE_NAME] = click_tag_lp['name']

    # Backup Asset Id
    backup_asset = self.creative_asset_dao.get(
        feed_item, column_name=FieldMap.CREATIVE_BACKUP_ASSET_ID)

    if backup_asset:
      feed_item[FieldMap.CREATIVE_BACKUP_ASSET_ID] = backup_asset['id']

    # Backup Landing Page Id
    backup_lp = self.landing_page_dao.get(
        feed_item,
        column_name=FieldMap.BACKUP_IMAGE_CLICK_THROUGH_LANDING_PAGE_ID)
    if backup_lp:
      feed_item[
          FieldMap.BACKUP_IMAGE_CLICK_THROUGH_LANDING_PAGE_ID] = backup_lp['id']
      feed_item[
          FieldMap
          .BACKUP_IMAGE_CLICK_THROUGH_LANDING_PAGE_NAME] = backup_lp['name']



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

# Field types for Schema Lookup
SDF_Field_Lookup = {
    # Campaign
    'Campaign Id': 'INTEGER',
    'Advertiser Id': 'INTEGER',
    'Name': 'STRING',
    'Timestamp': 'STRING',
    'Status': 'STRING',
    'Campaign Goal': 'STRING',
    'Campaign Goal KPI': 'STRING',
    'Creative Types': 'STRING',
    'Campaign Budget': 'FLOAT',
    'Campaign Start Date': 'STRING',
    'Campaign End Date': 'STRING',
    'Frequency Enabled': 'STRING',
    'Frequency Exposures': 'INTEGER',
    'Frequency Period': 'STRING',
    'Frequency Amount': 'INTEGER',
    'Demographic Targeting Gender': 'STRING',
    'Demographic Targeting Age': 'STRING',
    'Demographic Targeting Household Income': 'STRING',
    'Demographic Targeting Parental Status': 'STRING',
    'Geography Targeting - Include': 'STRING',
    'Geography Targeting - Exclude': 'STRING',
    'Language Targeting - Include': 'STRING',
    'Language Targeting - Exclude': 'STRING',
    'Brand Safety Labels': 'STRING',
    'Brand Safety Sensitivity Setting': 'STRING',
    'Brand Safety Custom Settings': 'STRING',
    'Third Party Services': 'STRING',
    'Third Party Verification Labels': 'STRING',
    'Viewability Targeting Active View': 'STRING',
    'Viewability Targeting Ad Position - Include': 'STRING',
    'Viewability Targeting Ad Position - Exclude': 'STRING',
    'Inventory Source Targeting - Authorized Seller Only': 'STRING',
    'Inventory Source Targeting - Include': 'STRING',
    'Inventory Source Targeting - Exclude': 'STRING',
    'Inventory Source Targeting - Target New Exchanges': 'STRING',
    'Environment Targeting': 'STRING',
    # Insertion Order
    'Io Id': 'INTEGER',
    'Io Type': 'STRING',
    'Fees': 'STRING',
    'Integration Code': 'STRING',
    'Details': 'STRING',
    'Pacing': 'STRING',
    'Pacing Rate': 'STRING',
    'Pacing Amount': 'FLOAT',
    'Performance Goal Type': 'STRING',
    'Performance Goal Value': 'STRING',
    'Measure DAR': 'STRING',
    'Measure DAR Channel': 'INTEGER',
    'Budget': 'STRING',
    'Budget Segments': 'STRING',
    'Auto Budget Allocation': 'STRING',
    'Device Targeting - Include': 'STRING',
    'Device Targeting - Exclude': 'STRING',
    'Browser Targeting - Include': 'STRING',
    'Browser Targeting - Exclude': 'STRING',
    'Brand Safety Labels': 'STRING',
    'Brand Safety Sensitivity Setting': 'STRING',
    'Brand Safety Custom Settings': 'STRING',
    'Third Party Verification Services': 'STRING',
    'Third Party Verification Labels': 'STRING',
    'Channel Targeting - Include': 'STRING',
    'Channel Targeting - Exclude': 'STRING',
    'Site Targeting - Include': 'STRING',
    'Site Targeting - Exclude': 'STRING',
    'App Targeting - Exclude': 'STRING',
    'App Collection Targeting - Include': 'STRING',
    'App Collection Targeting - Exclude': 'STRING',
    'Category Targeting - Include': 'STRING',
    'Category Targeting - Exclude': 'STRING',
    'Keyword Targeting - Include': 'STRING',
    'Keyword Targeting - Exclude': 'STRING',
    'Keyword List Targeting - Exclude': 'STRING',
    'Audience Targeting - Similar Audiences': 'STRING',
    'Audience Targeting - Include': 'STRING',
    'Audience Targeting - Exclude': 'STRING',
    'Affinity & In Market Targeting - Include': 'STRING',
    'Affinity & In Market Targeting - Exclude': 'STRING',
    'Custom List Targeting': 'STRING',
    'Inventory Source Targeting - Authorized Seller Only': 'STRING',
    'Inventory Source Targeting - Include': 'STRING',
    'Inventory Source Targeting - Exclude': 'STRING',
    'Inventory Source Targeting - Target New Exchanges': 'STRING',
    'Daypart Targeting': 'STRING',
    'Daypart Targeting Time Zone': 'STRING',
    'Environment Targeting': 'STRING',
    # Line Items
    'Line Item Id': 'INTEGER',
    'Budget Amount': 'FLOAT',
    'TrueView View Frequency Exposures': 'INTEGER',
    'Partner Revenue Amount': 'FLOAT',
    'Conversion Counting Pct': 'FLOAT',
    'Bid Strategy Value': 'FLOAT',
    'Bid Strategy Do Not Exceed': 'FLOAT',
    'TrueView Bid Strategy Value': 'FLOAT',
    'Bid Strategy Do Not Exceed': 'FLOAT',
    'TrueView Mobile Bid Adjustment Percentage': 'INTEGER',
    'TrueView Desktop Bid Adjustment Percentage': 'INTEGER',
    'TrueView Tablet Bid Adjustment Percentage': 'INTEGER',
    # Inventory Source
    'Rate': 'FLOAT',
    'Units Purchased': 'INTEGER',
    # Media Product
    'Product Id': 'INTEGER',
    'Plan Id': 'INTEGER',
    'Campaign Id': 'INTEGER',
    'Budget': 'FLOAT',
    'Frequency Exposures': 'INTEGER',
    'Frequency Amount': 'INTEGER',
    # Ad Group
    'Ad Group Id': 'INTEGER',
    'Line Item Id': 'INTEGER',
    'Bid Cost': 'FLOAT',
    'Popular Videos Bid Adjustment': 'INTEGER',
    # Ad
    'Ad Id': 'INTEGER',
    'Ad Group Id': 'INTEGER',
    'DCM Tracking - Placement Id': 'INTEGER',
    'DCM Tracking - Ad Id': 'INTEGER',
    'DCM Tracking - Creative Id': 'INTEGER'
}

SDF_COLUMNS = ('Line_Item_Id, Io_Id, Type, Subtype, Name, Timestamp, Status, '
               'Start_Date, End_Date, Budget_Type, Budget_Amount, Pacing, '
               'Pacing_Rate, Pacing_Amount, Frequency_Enabled, '
               'Frequency_Exposures, Frequency_Period, Frequency_Amount, '
               'Trueview_View_Frequency_Enabled, '
               'Trueview_View_Frequency_Exposures, '
               'Trueview_View_Frequency_Period, Partner_Revenue_Model, '
               'Partner_Revenue_Amount, Conversion_Counting_Type, '
               'Conversion_Counting_Pct, Conversion_Pixel_Ids, Fees, '
               'Integration_Code, Details, Bid_Strategy_Type, '
               'Bid_Strategy_Value, Bid_Strategy_Unit, '
               'Bid_Strategy_Do_Not_Exceed, Creative_Assignments, '
               'Geography_Targeting_Include, Geography_Targeting_Exclude, '
               'Language_Targeting_Include, Language_Targeting_Exclude, '
               'Device_Targeting_Include, Device_Targeting_Exclude, '
               'Browser_Targeting_Include, Browser_Targeting_Exclude, '
               'Brand_Safety_Labels, Brand_Safety_Sensitivity_Setting, '
               'Brand_Safety_Custom_Settings, '
               'Third_Party_Verification_Services, '
               'Third_Party_Verification_Labels, Channel_Targeting_Include, '
               'Channel_Targeting_Exclude, Site_Targeting_Include, '
               'Site_Targeting_Exclude, App_Targeting_Include, '
               'App_Targeting_Exclude, Category_Targeting_Include, '
               'Category_Targeting_Exclude, Keyword_Targeting_Include, '
               'Keyword_Targeting_Exclude, '
               'Audience_Targeting_Similar_Audiences, '
               'Audience_Targeting_Include, Audience_Targeting_Exclude, '
               'Affinity_In_Market_Targeting_Include, '
               'Affinity_In_Market_Targeting_Exclude, '
               'Custom_Affinity_Targeting, Inventory_Source_Targeting_Include, '
               'Inventory_Source_Targeting_Exclude, Daypart_Targeting, '
               'Environment_Targeting, Viewability_Targeting_Active_View, '
               'Viewability_Targeting_Ad_Position_Include, '
               'Viewability_Targeting_Ad_Position_Exclude, '
               'Video_Ad_Position_Targeting, Video_Player_Size_Targeting, '
               'Demographic_Targeting_Gender, Demographic_Targeting_Age, '
               'Connection_Speed_Targeting, Carrier_Targeting_Include, '
               'Carrier_Targeting_Exclude, '
               'Trueview_Mobile_Bid_Adjustment_Option, '
               'Trueview_Mobile_Bid_Adjustment_Percentage, '
               'Trueview_Category_Exclusions_Targeting, '
               'Trueview_Inventory_Source_Targeting')

SDF_FILE_TYPE_COLUMNS = {
    'AD': [
        'Line_Item_Id', 'Io_Id', 'Type', 'Subtype', 'Name', 'Timestamp',
        'Status', 'Start_Date', 'End_Date', 'Budget_Type', 'Budget_Amount',
        'Pacing', 'Pacing_Rate', 'Pacing_Amount', 'Frequency_Enabled',
        'Frequency_Exposures', 'Frequency_Period', 'Frequency_Amount',
        'Trueview_View_Frequency_Enabled', 'Trueview_View_Frequency_Exposures',
        'Trueview_View_Frequency_Period', 'Partner_Revenue_Model',
        'Partner_Revenue_Amount', 'Conversion_Counting_Type',
        'Conversion_Counting_Pct', 'Conversion_Pixel_Ids', 'Fees',
        'Integration_Code', 'Details', 'Bid_Strategy_Type',
        'Bid_Strategy_Value', 'Bid_Strategy_Unit', 'Bid_Strategy_Do_Not_Exceed',
        'Creative_Assignments', 'Geography_Targeting_Include',
        'Geography_Targeting_Exclude', 'Language_Targeting_Include',
        'Language_Targeting_Exclude', 'Device_Targeting_Include',
        'Device_Targeting_Exclude', 'Browser_Targeting_Include',
        'Browser_Targeting_Exclude', 'Brand_Safety_Labels',
        'Brand_Safety_Sensitivity_Setting', 'Brand_Safety_Custom_Settings',
        'Third_Party_Verification_Services', 'Third_Party_Verification_Labels',
        'Channel_Targeting_Include', 'Channel_Targeting_Exclude',
        'Site_Targeting_Include', 'Site_Targeting_Exclude',
        'App_Targeting_Include', 'App_Targeting_Exclude',
        'Category_Targeting_Include', 'Category_Targeting_Exclude',
        'Keyword_Targeting_Include', 'Keyword_Targeting_Exclude',
        'Audience_Targeting_Similar_Audiences', 'Audience_Targeting_Include',
        'Audience_Targeting_Exclude', 'Affinity_In_Market_Targeting_Include',
        'Affinity_In_Market_Targeting_Exclude', 'Custom_Affinity_Targeting',
        'Inventory_Source_Targeting_Include',
        'Inventory_Source_Targeting_Exclude', 'Daypart_Targeting',
        'Environment_Targeting', 'Viewability_Targeting_Active_View',
        'Viewability_Targeting_Ad_Position_Include',
        'Viewability_Targeting_Ad_Position_Exclude',
        'Video_Ad_Position_Targeting', 'Video_Player_Size_Targeting',
        'Demographic_Targeting_Gender', 'Demographic_Targeting_Age',
        'Connection_Speed_Targeting', 'Carrier_Targeting_Include',
        'Carrier_Targeting_Exclude', 'Trueview_Mobile_Bid_Adjustment_Option',
        'Trueview_Mobile_Bid_Adjustment_Percentage',
        'Trueview_Category_Exclusions_Targeting',
        'Trueview_Inventory_Source_Targeting'
    ],
    'AD_GROUP': [],
    'CAMPAIGN': [],
    'INSERTION_ORDER': [],
    'LINE_ITEM': [
        'Line_Item_Id', 'Io_Id', 'Type', 'Subtype', 'Name', 'Timestamp',
        'Status', 'Start_Date', 'End_Date', 'Budget_Type', 'Budget_Amount',
        'Pacing', 'Pacing_Rate', 'Pacing_Amount', 'Frequency_Enabled',
        'Frequency_Exposures', 'Frequency_Period', 'Frequency_Amount',
        'Trueview_View_Frequency_Enabled', 'Trueview_View_Frequency_Exposures',
        'Trueview_View_Frequency_Period', 'Partner_Revenue_Model',
        'Partner_Revenue_Amount', 'Conversion_Counting_Type',
        'Conversion_Counting_Pct', 'Conversion_Pixel_Ids', 'Fees',
        'Integration_Code', 'Details', 'Bid_Strategy_Type',
        'Bid_Strategy_Value', 'Bid_Strategy_Unit', 'Bid_Strategy_Do_Not_Exceed',
        'Creative_Assignments', 'Geography_Targeting_Include',
        'Geography_Targeting_Exclude', 'Language_Targeting_Include',
        'Language_Targeting_Exclude', 'Device_Targeting_Include',
        'Device_Targeting_Exclude', 'Browser_Targeting_Include',
        'Browser_Targeting_Exclude', 'Brand_Safety_Labels',
        'Brand_Safety_Sensitivity_Setting', 'Brand_Safety_Custom_Settings',
        'Third_Party_Verification_Services', 'Third_Party_Verification_Labels',
        'Channel_Targeting_Include', 'Channel_Targeting_Exclude',
        'Site_Targeting_Include', 'Site_Targeting_Exclude',
        'App_Targeting_Include', 'App_Targeting_Exclude',
        'Category_Targeting_Include', 'Category_Targeting_Exclude',
        'Keyword_Targeting_Include', 'Keyword_Targeting_Exclude',
        'Audience_Targeting_Similar_Audiences', 'Audience_Targeting_Include',
        'Audience_Targeting_Exclude', 'Affinity_In_Market_Targeting_Include',
        'Affinity_In_Market_Targeting_Exclude', 'Custom_Affinity_Targeting',
        'Inventory_Source_Targeting_Include',
        'Inventory_Source_Targeting_Exclude', 'Daypart_Targeting',
        'Environment_Targeting', 'Viewability_Targeting_Active_View',
        'Viewability_Targeting_Ad_Position_Include',
        'Viewability_Targeting_Ad_Position_Exclude',
        'Video_Ad_Position_Targeting', 'Video_Player_Size_Targeting',
        'Demographic_Targeting_Gender', 'Demographic_Targeting_Age',
        'Connection_Speed_Targeting', 'Carrier_Targeting_Include',
        'Carrier_Targeting_Exclude', 'Trueview_Mobile_Bid_Adjustment_Option',
        'Trueview_Mobile_Bid_Adjustment_Percentage',
        'Trueview_Category_Exclusions_Targeting',
        'Trueview_Inventory_Source_Targeting'
    ],
}



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import os

UI_ROOT = os.environ.get('STARTHINKER_ROOT', 'MISSING RUN deploy.sh TO SET')
UI_CRON = os.environ.get('STARTHINKER_CRON', '')

DEVELOPMENT_MODE = bool(os.environ.get('STARTHINKER_DEVELOPMENT', '0') == '1')

# used to manage user credentials and UI data
UI_ZONE = os.environ.get('STARTHINKER_ZONE', 'MISSING RUN deploy.sh TO SET')
UI_PROJECT = os.environ.get('STARTHINKER_PROJECT',
                            'MISSING RUN deploy.sh TO SET')
UI_SERVICE = os.environ.get('STARTHINKER_SERVICE',
                            'MISSING RUN deploy.sh TO SET')
UI_CLIENT = os.environ.get('STARTHINKER_CLIENT', 'MISSING RUN deploy.sh TO SET')

# used to multiply all buffer sizes for scaling on larger or smaller machines, can be a float
BUFFER_SCALE = int(os.environ.get('STARTHINKER_SCALE', 1))

# used to write execution trace when debugging
TRACE_FILE = '/tmp/starthinker_trace.log'

# used for user authentication
APPLICATION_NAME = 'StarThinker Client'
APPLICATION_SCOPES = [
    'https://www.googleapis.com/auth/userinfo.profile',
    'https://www.googleapis.com/auth/userinfo.email',
    'https://www.googleapis.com/auth/gmail.readonly',
    'https://www.googleapis.com/auth/gmail.send',
    'https://www.googleapis.com/auth/doubleclickbidmanager',
    'https://www.googleapis.com/auth/cloudplatformprojects',
    'https://www.googleapis.com/auth/devstorage.full_control',
    'https://www.googleapis.com/auth/bigquery',
    'https://www.googleapis.com/auth/dfareporting',
    'https://www.googleapis.com/auth/dfatrafficking',
    'https://www.googleapis.com/auth/drive',
    'https://www.googleapis.com/auth/spreadsheets',
    'https://www.googleapis.com/auth/doubleclicksearch',
    'https://www.googleapis.com/auth/content',
    'https://www.googleapis.com/auth/ddmconversions',
    'https://www.googleapis.com/auth/datastore',
    'https://www.googleapis.com/auth/logging.write',
    'https://www.googleapis.com/auth/logging.read',
    'https://www.googleapis.com/auth/pubsub',
    'https://www.googleapis.com/auth/youtube',
    'https://www.googleapis.com/auth/analytics',
    'https://www.googleapis.com/auth/analytics.readonly',
    'https://www.googleapis.com/auth/display-video',
    'https://www.googleapis.com/auth/adwords',
    'https://www.googleapis.com/auth/adsdatahub'
]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

# https://developers.google.com/bid-manager/v1/queries
# https://developers.google.com/drive/v3/web/manage-downloads
# https://developers.google.com/bid-manager/guides/entity-write/format

import re
import pprint
import json
import time
from types import GeneratorType
from urllib.request import urlopen

from starthinker.config import BUFFER_SCALE
from starthinker.util.project import project
from starthinker.util.data import get_rows
from starthinker.util.storage import object_get_chunks
from starthinker.util.csv import column_header_sanitize, csv_to_rows, rows_to_csv, response_utf8_stream
from starthinker.util.dbm.schema import LineItem_Write_Schema
from starthinker.util.google_api import API_DV360
from starthinker.util.google_api import API_DBM

DBM_CHUNKSIZE = int(
    200 * 1024000 *
    BUFFER_SCALE)  # 200MB recommended by docs * scale in config.py
RE_FILENAME = re.compile(r'.*/(.*)\?GoogleAccess')


def report_get(auth, report_id=None, name=None):
  """ Returns the DBM JSON definition of a report based on name or ID.

  Bulletproofing: https://developers.google.com/bid-manager/v1/queries/getquery

  Args:
    * auth: (string) Either user or service.
    * report_id: (int) ID of DCm report to fetch ( either or name ).
    * name: (string) Name of report to fetch ( either or report_id ).

  Returns:
    * JSON definition of report.

  """

  if name:
    for query in API_DBM(auth, iterate=True).queries().listqueries().execute():
      if query['metadata']['title'] == name:
        return query
  else:
    return API_DBM(auth).queries().getquery(queryId=report_id).execute()


def report_filter(auth, body, filters):
  """ Adds filters to a report body

  Filters cannot be easily added to the reports without templateing, this allows
  filters to be passed as lists.
  Values are specified using get_rows(...) helper, see
  starthinker/util/data/__init__.py.
  To specify a filter, use the official filter name and a list of values.

  For exmaple:

  ```
  filters = {
    "FILTER_PARTNER": {
      "values":789
    },
    "FILTER_ADVERTISER": {
      "values":[1234, 5678, 91011]
    }
  }
  ```

  Args:
    * auth: (string) Either user or service.
    * body: (json) the report body ( with or without filters )
    * filters: (json) a dictionary of filters to apply ( see above examples )

  Returns:
    * body: ( json ) modified report body
  """

  new_body = body.copy()

  for f, d in filters.items():
    for v in get_rows(project.task['auth'], d):
      new_body['params'].setdefault('filters', []).append({
          'type': f,
          'value': v
      })

  return new_body


def report_build(auth, body):
  """ Creates a DBM report given a JSON definition.

  Bulletproofing:
  https://developers.google.com/bid-manager/v1/queries/createquery

  The report will be automatically run the first time.

  The body JSON provided will have the following fields added if not present:
    * schedule - set to run daily and expire in one year.

  Args:
    * auth: (string) Either user or service.
    * body: (json) As defined in:
      https://developers.google.com/doubleclick-advertisers/v3.2/reports#resource

  Returns:
    * JSON definition of report created or existing.

  """

  report = report_get(auth, name=body['metadata']['title'])

  if not report:

    # add default daily schedule if it does not exist ( convenience )
    body.setdefault('schedule', {})
    body['schedule'].setdefault('nextRunTimezoneCode', body['timezoneCode'])
    body['schedule'].setdefault('frequency', 'DAILY')
    if body['schedule']['frequency'] != 'ONE_TIME':
      body['schedule'].setdefault('nextRunMinuteOfDay', 4 * 60)
      body['schedule'].setdefault('endTimeMs',
                                  int((time.time() + (365 * 24 * 60 * 60)) *
                                      1000))  # 1 year in future

    # build report
    #pprint.PrettyPrinter().pprint(body)
    report = API_DBM(auth).queries().createquery(body=body).execute()

    # run report first time
    body = {
        'dataRange': report['metadata']['dataRange'],
        'timezoneCode': body['timezoneCode']
    }
    API_DBM(auth).queries().runquery(
        queryId=report['queryId'], body=body).execute()

  else:
    if project.verbose:
      print('DBM Report Exists:', body['metadata']['title'])

  return report


def report_fetch(auth, report_id=None, name=None, timeout=60):
  """ Retrieves most recent DBM file JSON by name or ID, if in progress, waits for it to complete.

  Bulletproofing: https://developers.google.com/bid-manager/v1/queries/getquery

  Timeout is in minutes ( retries will happen at 5 minute interval, default
  total time is 20 minutes )

  Args:
    * auth: (string) Either user or service.
    * report_id: (int) ID of DCm report to fetch ( either or name ).
    * name: (string) Name of report to fetch ( either or report_id ).
    * timeout: (int) Minutes to wait for in progress report before giving up.

  Returns:
    * Report JSON if report exists and is ready.
    * True if report is in progress but not ready.
    * False if report does not exist.

  """

  if project.verbose:
    print('DBM Report Download ( timeout ):', report_id or name, timeout)

  while timeout >= 0:  # allow zero to execute at least once
    # advance timeout first ( if = 0 then exit condition met but already in loop, if > 0 then will run into sleep )
    timeout -= 1

    report = report_get(auth, report_id, name)
    #pprint.PrettyPrinter().pprint(report)
    if report:
      # report is running ( return only if timeout is exhausted )
      if report['metadata'].get('googleCloudStoragePathForLatestReport',
                                '') == '':
        if project.verbose:
          print('DBM Still Running')
        if timeout < 0:
          return True
      # file exists ( return it success )
      else:
        return report['metadata']['googleCloudStoragePathForLatestReport']

    # no report ( break out of loop it will never finish )
    else:
      if project.verbose:
        print('DBM No Report')
      return False

    # sleep a minute
    if timeout > 0:
      if project.verbose:
        print('WAITING MINUTES', timeout)
      time.sleep(60)


def report_file(auth,
                report_id=None,
                name=None,
                timeout=60,
                chunksize=DBM_CHUNKSIZE):
  """ Retrieves most recent DBM file by name or ID, if in progress, waits for it to complete.

  Bulletproofing: https://developers.google.com/bid-manager/v1/queries/getquery

  Timeout is in minutes ( retries will happen at 1 minute interval, default
  total time is 60 minutes )
  If chunksize is set to None then the whole file is downloaded at once.

  Args:
    * auth: (string) Either user or service.
    * report_id: (int) ID of DCm report to fetch ( either or name ).
    * name: (string) Name of report to fetch ( either or report_id ).
    * timeout: (int) Minutes to wait for in progress report before giving up.
    * chunksize: (int) number of bytes to download at a time, for memory
      constrained systems.

  Returns:
    * (filename, iterator) if file exists and is ready to download in chunks.
    * (filename, file) if file exists and chunking is off.
    * ('report_running.csv', None) if report is in progress.
    * (None, None) if file does not exist.

  """

  storage_path = report_fetch(auth, report_id, name, timeout)

  if storage_path == False:
    return None, None
  elif storage_path == True:
    return 'report_running.csv', None
  else:
    filename = RE_FILENAME.search(storage_path).groups(0)[0]

    # streaming
    if chunksize:
      if project.verbose:
        print('REPORT FILE STREAM:', storage_path)
      return filename, response_utf8_stream(urlopen(storage_path), chunksize)

    # single object
    else:
      if project.verbose:
        print('REPORT FILE SINGLE:', storage_path)
      return filename, urlopen(storage_path).read().decode('UTF-8')


def report_delete(auth, report_id=None, name=None):
  """ Deletes a DBM report based on name or ID.

  Bulletproofing:
  https://developers.google.com/bid-manager/v1/queries/deletequery

  Args:
    * auth: (string) Either user or service.
    * report_id: (int) ID of DCm report to fetch ( either or name ).
    * name: (string) Name of report to fetch ( either or report_id ).

  Returns:
    * None

  """

  if project.verbose:
    print('DBM DELETE:', report_id or name)
  report = report_get(auth, report_id, name)
  if report:
    API_DBM(auth).queries().deletequery(queryId=report['queryId']).execute()
  else:
    if project.verbose:
      print('DBM DELETE: No Report')


def report_list(auth):
  """ Lists all the DBM report configurations for the current credentials.

  Bulletproofing:
  https://developers.google.com/bid-manager/v1/queries/listqueries

  Args:
    * auth: (string) Either user or service.

  Returns:
    * Iterator of JSONs.

  """

  for query in API_DBM(auth, iterate=True).queries().listqueries().execute():
    yield query


""" Get a DV360 report as a list

Args: * auth => auth from the job * report_id => the report id that wants to be
pulled

Returns:
  * a DV360 report represented as a list
"""


def report_to_list(auth, report_id):
  filename, report = report_file(
      auth,
      report_id,
      None,  #name
      10,  #timeout
      DBM_CHUNKSIZE)

  if report:
    rows = report_to_rows(report)
    rows = report_clean(rows)

  return list(rows)


def report_to_rows(report):
  """ Helper to convert DBM files into iterator of rows, memory efficient.

  Usage example:

  ```
  filename, report = report_file(...)
  rows = report_to_rows(report)
  ```

  Args:
    * report: (iterator or file) Either an iterator or file that will be
      converted to rows.

  Returns:
    * Iterator of lists representing each row.

  """

  # if reading from stream
  if type(report) is GeneratorType:
    leftovers = ''
    for chunk in report:
      data, extra = chunk.rsplit('\n', 1)
      for row in csv_to_rows(leftovers + data):
        yield row
      leftovers = extra

  # if reading from buffer
  else:
    for row in csv_to_rows(report):
      yield row


def report_clean(rows):
  """ Helper to fix DBM report issues for BigQuery and ensure schema compliance.

  Memory efficiently cleans each row by fixing:
  * Strips header and footer to preserve only data rows.
  * Changes 'Date' to 'Report_Day' to avoid using reserved name in BigQuery.
  * Changes date values to use '-' instead of '/' for BigQuery compatibility.
  * Changes columns '-' and 'Unknown' to NULL
  * Changes '< 1000' to 1000

  Usage example:

  ```
  filename, report = report_file(...)
  rows = report_to_rows(report)
  rows = report_clean(rows)
  ```

  Args:
    * rows: (iterator) Rows to clean.

  Returns:
    * Iterator of cleaned rows.

  """

  if project.verbose:
    print('DBM Report Clean')

  first = True
  last = False
  date = None
  for row in rows:
    # stop if no data returned
    if row == ['No data returned by the reporting service.']:
      break

    # stop at blank row ( including sum row )
    if not row or row[0] is None or row[0] == '':
      break

    # sanitizie header row
    if first:
      try:
        date_column = row.index('Date')
        row[date_column] = 'Report_Day'
      except ValueError:
        pass
      row = [column_header_sanitize(cell) for cell in row]

    # for all data rows clean up cells
    else:
      # check if data studio formatting is applied reformat the dates
      row = [
          cell.replace('/', '-') if isinstance(cell, str) and len(cell) == 4 +
          1 + 2 + 1 + 2 and cell[4] == '/' and cell[7] == '/' else cell
          for cell in row
      ]  # 5x faster than regexp

    # remove unknown columns ( which throw off schema on import types )
    row = [
        '' if cell.strip() in (
            'Unknown',
            '-',
        ) else ('1000' if cell == '< 1000' else cell) for cell in row
    ]

    # return the row
    yield row

    # not first row anymore
    first = False


def lineitem_patch_v1(auth, patch, li):
  """Patches a DV360 Line Item

  Args:
    auth: StarThinker authentication scheme
    patch: List of field names to patch
    li: Line item with updates to push
  Returns: Updated Line Item
  """
  return API_DV360(auth).advertisers().lineItems().patch(
      advertiserId=li['advertiserId'],
      lineItemId=li['lineItemId'],
      updateMask=patch,
      body=li).execute()


def lineitem_get_v1(auth, advertiser_id, lineitem_id):
  """Gets a DV360 Line Item

  Args:
    auth: StarThinker authentication scheme
    advertiser_id: ID of the advertiser of the line item
    lineitem_id: ID of the line item
  Returns: Line Item from the DV360 API
  """
  return API_DV360(auth).advertisers().lineItems().get(
      advertiserId=advertiser_id, lineItemId=lineitem_id).execute()


def lineitem_read(auth, advertisers=[], insertion_orders=[], lineitems=[]):
  """ Reads line item configurations from DBM.

  Bulletproofing:
  https://developers.google.com/bid-manager/v1/lineitems/downloadlineitems

  Args:
    * auth: (string) Either user or service. * advertisers (list) List of
      advertiser ids ( exclusive with insertion_orders and lineitems ). *
      insertion_orders (list) List of insertion_order ids ( exclusive with
      advertisers and lineitems ). * lineitems (list) List of ilineitem ids (
      exclusive with insertion_orders and advertisers ).

  Returns:
    * Iterator of lists:
    https://developers.google.com/bid-manager/guides/entity-write/format

  """

  body = {'format': 'CSV', 'fileSpec': 'EWF'}

  if advertisers:
    body['filterType'] = 'ADVERTISER_ID'
    body['filterIds'] = list(advertisers)  # in case its a generator

  elif insertion_orders:
    body['filterType'] = 'INSERTION_ORDER_ID'
    body['filterIds'] = list(insertion_orders)  # in case its a generator

  elif lineitems:
    body['filterType'] = 'LINE_ITEM_ID'
    body['filterIds'] = list(lineitems)  # in case its a generator

  #print(body)

  result = API_DBM(auth).lineitems().downloadlineitems(body=body).execute()

  for count, row in enumerate(csv_to_rows(result.get('lineItems', ''))):
    if count == 0:
      continue  # skip header
    row[0] = int(row[0] or 0)  # LineItem ID
    row[2] = int(row[2] or 0)  # Partner ID
    row[11] = float(row[11] or 0)  # IO Budget Amount
    row[18] = float(row[18] or 0)  # LineItem Budget Amount
    row[21] = float(row[21] or 0)  # LineItem Pacing Amount
    row[23] = int(row[23] or 0)  # LineItem Frequency Exposures
    row[25] = int(row[25] or 0)  # LineItem Frequency Amount
    row[26] = float(row[26] or 0)  # Bid Price (CPM)
    row[28] = float(row[28] or 0)  # Partner Revenue Amount
    yield row


def lineitem_edit(row, column_name, value):
  lineitem_lookup = {k['name']: v for v, k in enumerate(LineItem_Write_Schema)}
  row[lineitem_lookup[column_name]] = value


def lineitem_write(auth, rows, dry_run=True):
  """ Writes a list of lineitem configurations to DBM.

  Bulletproofing:
  https://developers.google.com/bid-manager/v1/lineitems/uploadlineitems

   Args:
    * auth: (string) Either user or service.
    * rows (iterator) List of lineitems:
      https://developers.google.com/bid-manager/guides/entity-write/format *
      dry_run (boolean) If set to True no write will occur, only a test of the
      upload for errors.

  Returns:
    * Results of upload.

  """

  header = [s['name'] for s in LineItem_Write_Schema]

  body = {
      'lineItems': '%s\n%s' % (','.join(header),
                               rows_to_csv(rows).read()),  # add header row
      'format': 'CSV',
      'dryRun': dry_run
  }

  result = API_DBM(auth).lineitems().uploadlineitems(body=body).execute()
  #print(result)
  return (result)


def sdf_read(auth, file_types, filter_type, filter_ids, version='3.1'):
  """ Read sdf file from DV360 api

  https://developers.google.com/bid-manager/v1/sdf/download

  Args:
    * auth:  (string) Either user or service.
    * file_types: (list[string]) List of the requested file types
    * filter_type: (string) Value of the type of filter ids to be expected
    * filter_ids: (list[int]) Filter ids for the api call

  Returns:
    * Rows of the sdf files requested

  """

  body = {
      'fileTypes': file_types,
      'filterType': filter_type,
      'filterIds': filter_ids,
      'version': version
  }

  result = API_DBM(auth).sdf().download(body=body).execute()

  # Clean the date field in each row
  for key in result.keys():
    date_idx_list = []
    for count, row in enumerate(csv_to_rows(result.get(key, ''))):
      # Get date fields from header
      if count == 0:
        date_idx_list = _get_idx_sdf_date_fields(row)
        yield row
      # Clean and yield data rows
      else:
        yield _clean_sdf_row(row, date_idx_list)


def _get_idx_sdf_date_fields(header):
  date_idx_list = []

  for idx, header in enumerate(header):
    if 'date' in header or 'Date' in header:
      date_idx_list.append(idx)

  return date_idx_list


def _clean_sdf_row(row, date_idx_list):
  # fix format: MM/DD/YYYY HH:MM   =>   YYYY-MM-DD HH:MM
  for idx in date_idx_list:
    row[idx] = _convert_sdf_date(row[idx])

  return row


def _convert_sdf_date(date_string):
  return date_string[6:9] + '-' + date_string[:1] + '-' + date_string[
      3:4] + ' ' + date_string[11:12] + ':' + date_string[14:15]



from datetime import date
import json
import argparse

from starthinker.util.project import project
from starthinker.util.google_api import API_SNIPPETS

if __name__ == '__main__':

  # get parameters
  parser = argparse.ArgumentParser()
  parser.add_argument(
      '--ldap', '-l', help='Name of user to fetch snippets for.', default=None)
  parser.add_argument(
      '--weeks_back',
      '-w',
      help='Number of weeks to go back, 0 is this week.',
      default=0)

  # initialize project
  project.from_commandline(parser=parser)

  results = API_SNIPPETS('user').listByConstraints(
      startDate='09-03-2108',  #date.today().strftime('%m-%d-%Y'),
      #endDate=date.today().strftime('%m-%d-%Y'),
      userGroupKey='kenjora',  #project.args.ldap,
      fields='items(text)').execute()

  print(results)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re
import sys
import codecs
import csv
import uuid
import json
from time import sleep
from io import BytesIO
from googleapiclient.errors import HttpError
from googleapiclient.http import MediaIoBaseUpload
from google.cloud.bigquery._helpers import _row_tuple_from_json

from starthinker.config import BUFFER_SCALE
from starthinker.util import flag_last
from starthinker.util.project import project
from starthinker.util.google_api import API_BigQuery, API_Retry
from starthinker.util.csv import row_header_sanitize

BIGQUERY_BUFFERMAX = 4294967296
BIGQUERY_CHUNKSIZE = int(200 * 1024000 *
                         BUFFER_SCALE)  # 200 MB * scale in config.py
BIGQUERY_BUFFERSIZE = min(BIGQUERY_CHUNKSIZE * 4,
                          BIGQUERY_BUFFERMAX)  # 1 GB * scale in config.py

RE_TABLE_NAME = re.compile(r'[^\w]+')
RE_INDENT = re.compile(r' {5,}')


def row_to_json(row, schema, as_object=False):

  if as_object:
    row_raw = {'f': [{'v': row}]}
    schema_raw = [{
        'name': 'wrapper',
        'type': 'RECORD',
        'mode': 'REQUIRED',
        'fields': schema
    }]
    return _row_tuple_from_json(row_raw, schema_raw)[0]

  else:
    row_raw = row
    schema_raw = schema
    return list(_row_tuple_from_json(row_raw, schema_raw))


def bigquery_date(value):
  return value.strftime('%Y%m%d')


def table_name_sanitize(name):
  return RE_TABLE_NAME.sub('_', name)


def query_parameters(query, parameters):
  """Replace variables in a query string with values.

  CAUTION: Possible SQL injection, please check up stream.
  query = "SELECT * FROM {project}.{dataset}.Some_Table"
  parameters = {'project': 'Test_Project', 'dataset':'Test_dataset'}
  print query_parameters(query, parameters)
  """

  # no effect other than visual formatting
  query = RE_INDENT.sub(r'\n\g<0>', query)

  if not parameters:
    return query
  elif isinstance(parameters, dict):
    return query.format(**parameters)
  else:
    while '[PARAMETER]' in query:
      try:
        parameter = parameters.pop(0)
      except IndexError:
        raise IndexError('BigQuery: Missing PARAMETER values for this query.')
      if isinstance(parameter, list) or isinstance(parameter, tuple):
        parameter = ', '.join([str(p) for p in parameter])
      query = query.replace('[PARAMETER]', parameter, 1)
    if project.verbose:
      print('QUERY:', query)
    return query


def job_wait(auth, job):
  if job:
    if project.verbose:
      print('BIGQUERY JOB WAIT:', job['jobReference']['jobId'])

    request = API_BigQuery(auth).jobs().get(
        projectId=job['jobReference']['projectId'],
        jobId=job['jobReference']['jobId'])

    while True:
      sleep(5)
      if project.verbose:
        print('.', end='')
      sys.stdout.flush()
      result = API_Retry(request)
      if 'errors' in result['status']:
        raise Exception(
            'BigQuery Job Error: %s' %
            ' '.join([e['message'] for e in result['status']['errors']]))
      elif 'errorResult' in result['status']:
        raise Exception('BigQuery Job Error: %s' %
                        result['status']['errorResult']['message'])
      elif result['status']['state'] == 'DONE':
        if project.verbose:
          print('JOB COMPLETE:', result['id'])
        break


def datasets_create(auth, project_id, dataset_id):

  body = {
      'description': dataset_id,
      'datasetReference': {
          'projectId': project_id,
          'datasetId': dataset_id,
      },
      'location': 'US',
      'friendlyName': dataset_id,
  }

  API_BigQuery(auth).datasets().insert(
      projectId=project_id, body=body).execute()


def datasets_delete(auth, project_id, dataset_id, delete_contents=True):

  try:
    API_BigQuery(auth).datasets().delete(
      projectId=project_id,
      datasetId=dataset_id,
      deleteContents=delete_contents
    ).execute()
    return True
  except HttpError as e:
    if e.resp.status != 404:
      raise
    return False


# roles = READER, WRITER, OWNER
def datasets_access(auth,
                    project_id,
                    dataset_id,
                    role='READER',
                    emails=[],
                    groups=[],
                    views=[]):

  if emails or groups or views:
    access = API_BigQuery(auth).datasets().get(
        projectId=project_id, datasetId=dataset_id).execute()['access']

    # if emails
    for email in emails:
      access.append({
          'userByEmail': email,
          'role': role,
      })

    # if groups
    for group in groups:
      access.append({
          'groupByEmail': group,
          'role': role,
      })

    for view in views:
      access.append({
          'view': {
              'projectId': project_id,
              'datasetId': view['dataset'],
              'tableId': view['view']
          }
      })

    API_BigQuery(auth).datasets().patch(
        projectId=project_id, datasetId=dataset_id, body={
            'access': access
        }).execute()


def run_query(auth, project_id, query, legacy=True, dataset_id=None):

  body = {'query': query, 'useLegacySql': legacy}

  if dataset_id:
    body['defaultDataset'] = {'datasetId': dataset_id}

  job_wait(
      auth,
      API_BigQuery(auth).jobs().query(projectId=project_id,
                                      body=body).execute())


def query_to_table(auth,
                   project_id,
                   dataset_id,
                   table_id,
                   query,
                   disposition='WRITE_TRUNCATE',
                   legacy=True,
                   billing_project_id=None,
                   target_project_id=None):
  target_project_id = target_project_id or project_id

  if not billing_project_id:
    billing_project_id = project_id

  body = {
      'configuration': {
          'query': {
              'useLegacySql': legacy,
              'query': query,
              'destinationTable': {
                  'projectId': target_project_id,
                  'datasetId': dataset_id,
                  'tableId': table_id
              },
              'createDisposition': 'CREATE_IF_NEEDED',
              'writeDisposition': disposition,
              'allowLargeResults': True
          },
      }
  }

  job_wait(
      auth,
      API_BigQuery(auth).jobs().insert(projectId=billing_project_id,
                                       body=body).execute())


def query_to_view(auth,
                  project_id,
                  dataset_id,
                  view_id,
                  query,
                  legacy=True,
                  replace=False):

  body = {
      'tableReference': {
          'projectId': project_id,
          'datasetId': dataset_id,
          'tableId': view_id,
      },
      'view': {
          'query': query,
          'useLegacySql': legacy
      }
  }

  response = API_BigQuery(auth).tables().insert(
      projectId=project_id, datasetId=dataset_id, body=body).execute()
  if response is None and replace:
    return API_BigQuery(auth).tables().update(
        projectId=project_id, datasetId=dataset_id, tableId=view_id,
        body=body).execute()


def storage_to_table(auth,
                     project_id,
                     dataset_id,
                     table_id,
                     path,
                     schema=[],
                     skip_rows=1,
                     structure='CSV',
                     disposition='WRITE_TRUNCATE',
                     wait=True):
  if project.verbose:
    print('BIGQUERY STORAGE TO TABLE: ', project_id, dataset_id, table_id)

  body = {
      'configuration': {
          'load': {
              'destinationTable': {
                  'projectId': project_id,
                  'datasetId': dataset_id,
                  'tableId': table_id,
              },
              'sourceFormat': 'NEWLINE_DELIMITED_JSON',
              'writeDisposition':
                  disposition,  # WRITE_TRUNCATE, WRITE_APPEND, WRITE_EMPTY
              'autodetect': True,
              'allowJaggedRows': True,
              'allowQuotedNewlines': True,
              'ignoreUnknownValues': True,
              'sourceUris': ['gs://%s' % path.replace(':', '/'),],
          }
      }
  }

  if schema:
    body['configuration']['load']['schema'] = {'fields': schema}
    body['configuration']['load']['autodetect'] = False

  if structure == 'CSV':  # CSV, NEWLINE_DELIMITED_JSON
    body['configuration']['load']['sourceFormat'] = 'CSV'
    body['configuration']['load']['skipLeadingRows'] = skip_rows

  job = API_BigQuery(auth).jobs().insert(
      projectId=project_id, body=body).execute()
  if wait:
    try:
      job_wait(auth, job)
    except Exception as e:
      print('BIGQUERY SKIPPING: %s, %s' % (path, str(e)))
  else:
    return job


def rows_to_table(auth,
                  project_id,
                  dataset_id,
                  table_id,
                  rows,
                  schema=[],
                  skip_rows=1,
                  disposition='WRITE_TRUNCATE',
                  wait=True):
  if project.verbose:
    print('BIGQUERY ROWS TO TABLE: ', project_id, dataset_id, table_id)

  buffer_data = BytesIO()
  buffer_writer = codecs.getwriter('utf-8')
  writer = csv.writer(
      buffer_writer(buffer_data),
      delimiter=',',
      quotechar='"',
      quoting=csv.QUOTE_MINIMAL)
  has_rows = False

  if rows == []:
    if project.verbose:
      print('BigQuery Zero Rows')
    return io_to_table(auth, project_id, dataset_id, table_id, buffer_data,
                       'CSV', schema, skip_rows, disposition, wait)

  for is_last, row in flag_last(rows):

    # write row to csv buffer
    writer.writerow(row)

    # write the buffer in chunks
    if is_last or buffer_data.tell() + 1 > BIGQUERY_BUFFERSIZE:
      if project.verbose:
        print('BigQuery Buffer Size', buffer_data.tell())
      buffer_data.seek(0)  # reset for read
      io_to_table(auth, project_id, dataset_id, table_id, buffer_data, 'CSV',
                  schema, skip_rows, disposition)

      # reset buffer for next loop, be sure to do an append to the table
      buffer_data.seek(0)  #reset for write
      buffer_data.truncate(
      )  # reset for write ( yes its needed for EOF marker )
      disposition = 'WRITE_APPEND'  # append all remaining records
      skip_rows = 0
      has_rows = True

  # if no rows, clear table to simulate empty write
  if not has_rows:
    if project.verbose:
      print('BigQuery Zero Rows')
    return io_to_table(auth, project_id, dataset_id, table_id, buffer_data,
                       'CSV', schema, skip_rows, disposition, wait)


def json_to_table(auth,
                  project_id,
                  dataset_id,
                  table_id,
                  json_data,
                  schema=None,
                  disposition='WRITE_TRUNCATE',
                  wait=True):
  if project.verbose:
    print('BIGQUERY JSON TO TABLE: ', project_id, dataset_id, table_id)

  buffer_data = BytesIO()
  has_rows = False

  for is_last, record in flag_last(json_data):

    # check if json is already string encoded, and write to buffer
    buffer_data.write((record if isinstance(record, str) else
                       json.dumps(record)).encode('utf-8'))

    # write the buffer in chunks
    if is_last or buffer_data.tell() + 1 > BIGQUERY_BUFFERSIZE:
      if project.verbose:
        print('BigQuery Buffer Size', buffer_data.tell())
      buffer_data.seek(0)  # reset for read
      io_to_table(auth, project_id, dataset_id, table_id, buffer_data,
                  'NEWLINE_DELIMITED_JSON', schema, 0, disposition)

      # reset buffer for next loop, be sure to do an append to the table
      buffer_data.seek(0)  #reset for write
      buffer_data.truncate(
      )  # reset for write ( yes its needed for EOF marker )
      disposition = 'WRITE_APPEND'  # append all remaining records
      has_rows = True

    # if not end append newline, for newline delimited json
    else:
      buffer_data.write('\n'.encode('utf-8'))

  # if no rows, clear table to simulate empty write
  if not has_rows:
    if project.verbose:
      print('BigQuery Zero Rows')
    return io_to_table(auth, project_id, dataset_id, table_id, buffer_data,
                       'NEWLINE_DELIMITED_JSON', schema, 0, disposition, wait)


def io_to_table(auth,
                project_id,
                dataset_id,
                table_id,
                data_bytes,
                source_format='CSV',
                schema=None,
                skip_rows=0,
                disposition='WRITE_TRUNCATE',
                wait=True):

  # if data exists, write data to table
  data_bytes.seek(0, 2)
  if data_bytes.tell() > 0:
    data_bytes.seek(0)

    media = MediaIoBaseUpload(
        data_bytes,
        mimetype='application/octet-stream',
        resumable=True,
        chunksize=BIGQUERY_CHUNKSIZE)

    body = {
        'configuration': {
            'load': {
                'destinationTable': {
                    'projectId': project_id,
                    'datasetId': dataset_id,
                    'tableId': table_id,
                },
                'sourceFormat': source_format,  # CSV, NEWLINE_DELIMITED_JSON
                'writeDisposition':
                    disposition,  # WRITE_TRUNCATE, WRITE_APPEND, WRITE_EMPTY
                'autodetect': True,
                'allowJaggedRows': True,
                'allowQuotedNewlines': True,
                'ignoreUnknownValues': True,
            }
        }
    }

    if schema:
      body['configuration']['load']['schema'] = {'fields': schema}
      body['configuration']['load']['autodetect'] = False

    if disposition == 'WRITE_APPEND':
      body['configuration']['load']['autodetect'] = False

    if source_format == 'CSV':
      body['configuration']['load']['skipLeadingRows'] = skip_rows

    job = API_BigQuery(auth).jobs().insert(
        projectId=project.id, body=body, media_body=media).execute(run=False)
    execution = job.execute()

    response = None
    while response is None:
      status, response = job.next_chunk()
      if project.verbose and status:
        print('Uploaded %d%%.' % int(status.progress() * 100))
    if project.verbose:
      print('Uploaded 100%')

    if wait:
      job_wait(auth, execution)
    else:
      return execution

  # if it does not exist and write, clear the table
  elif disposition == 'WRITE_TRUNCATE':
    if project.verbose:
      print('BIGQUERY: No data, clearing table.')
    table_create(auth, project_id, dataset_id, table_id, schema)


def incremental_rows_to_table(auth,
                              project_id,
                              dataset_id,
                              table_id,
                              rows,
                              schema=[],
                              skip_rows=1,
                              disposition='WRITE_APPEND',
                              billing_project_id=None):
  if project.verbose:
    print('BIGQUERY INCREMENTAL ROWS TO TABLE: ', project_id, dataset_id,
          table_id)

  #load the data in rows to BQ into a temp table
  table_id_temp = table_id + str(uuid.uuid4()).replace('-', '_')
  rows_to_table(auth, project_id, dataset_id, table_id_temp, rows, schema,
                skip_rows, disposition)

  try:
    #query the temp table to find the max and min date
    start_date = _get_min_date_from_table(
        auth,
        project_id,
        dataset_id,
        table_id_temp,
        billing_project_id=billing_project_id)
    end_date = _get_max_date_from_table(
        auth,
        project_id,
        dataset_id,
        table_id_temp,
        billing_project_id=billing_project_id)

    #check if master table exists: if not create it, if so clear old data
    if not table_exists(auth, project_id, dataset_id, table_id):
      table_create(auth, project_id, dataset_id, table_id)
    else:
      _clear_data_in_date_range_from_table(
          auth,
          project_id,
          dataset_id,
          table_id,
          start_date,
          end_date,
          billing_project_id=billing_project_id)

    #append temp table to master
    query = ('SELECT * FROM `' + project_id + '.' + dataset_id + '.' +
             table_id_temp + '` ')
    query_to_table(
        auth,
        project_id,
        dataset_id,
        table_id,
        query,
        disposition,
        False,
        billing_project_id=billing_project_id)

    #delete temp table
    drop_table(
        auth,
        project_id,
        dataset_id,
        table_id_temp,
        billing_project_id=billing_project_id)

  except:
    #delete temp table
    drop_table(
        auth,
        project_id,
        dataset_id,
        table_id_temp,
        billing_project_id=billing_project_id)


def table_create(auth,
                 project_id,
                 dataset_id,
                 table_id,
                 schema=None,
                 overwrite=True,
                 is_time_partition=False):

  if overwrite:
    table_delete(auth, project_id, dataset_id, table_id)

  body = {
      'tableReference': {
          'projectId': project_id,
          'tableId': table_id,
          'datasetId': dataset_id,
      }
  }

  if schema:
    body['schema'] = {'fields': schema}

  if is_time_partition:
    body['timePartitioning'] = {'type': 'DAY'}

  API_BigQuery(auth).tables().insert(
      projectId=project_id, datasetId=dataset_id, body=body).execute()


def table_get(auth, project_id, dataset_id, table_id):
  return API_BigQuery(auth).tables().get(
      projectId=project_id, datasetId=dataset_id, tableId=table_id).execute()


def table_list(auth, project_id, dataset_id=None):
  if dataset_id is None:
    datasets = [
      d['datasetReference']['datasetId'] for d in API_BigQuery(auth).datasets().list(
        projectId=project_id,
        fields='datasets.datasetReference.datasetId, nextPageToken'
      ).execute()
    ]
  else:
    datasets = [dataset_id]

  for dataset_id in datasets:
    for table in API_BigQuery(auth, iterate=True).tables().list(
        projectId=project_id,
        datasetId=dataset_id,
        fields='tables.tableReference, tables.type, nextPageToken'
    ).execute():
      yield table['tableReference']['datasetId'], table['tableReference']['tableId'], table['type']


def table_exists(auth, project_id, dataset_id, table_id):
  try:
    table_get(auth, project_id, dataset_id, table_id)
    return True
  except HttpError as e:
    if e.resp.status != 404:
      raise
    return False


def table_delete(auth, project_id, dataset_id, table_id):
  try:
    API_BigQuery(auth).tables().delete(
        projectId=project_id, datasetId=dataset_id, tableId=table_id).execute()
    return True
  except HttpError as e:
    if e.resp.status != 404:
      raise
    return False



def table_copy(auth, from_project, from_dataset, from_table, to_project,
               to_dataset, to_table):

  body = {
      'copy': {
          'sourceTable': {
              'projectId': from_project,
              'datasetId': from_dataset,
              'tableId': from_table
          },
          'destinationTable': {
              'projectId': to_project,
              'datasetId': to_dataset,
              'tableId': to_table
          }
      }
  }

  job_wait(
      auth,
      API_BigQuery(auth).jobs().insert(projectId=project.id,
                                       body=body).execute())


def table_to_rows(auth,
                  project_id,
                  dataset_id,
                  table_id,
                  fields=None,
                  row_start=0,
                  row_max=None,
                  as_object=False):
  if project.verbose:
    print('BIGQUERY ROWS:', project_id, dataset_id, table_id)

  table = API_BigQuery(auth).tables().get(
      projectId=project_id, datasetId=dataset_id, tableId=table_id).execute()
  table_schema = table['schema'].get('fields', [])
  table_type = table['type']
  table_legacy = table.get('view', {}).get('useLegacySql', False)

  if table_type == 'TABLE':

    for row in API_BigQuery(
        auth, iterate=True).tabledata().list(
            projectId=project_id,
            datasetId=dataset_id,
            tableId=table_id,
            selectedFields=fields,
            startIndex=row_start,
            maxResults=row_max,
        ).execute():
      yield row_to_json(row, table_schema, as_object)

  else:
    yield from query_to_rows(auth, project_id, dataset_id,
                             'SELECT * FROM %s' % table_id, row_max,
                             table_legacy, as_object)


def table_to_schema(auth, project_id, dataset_id, table_id):
  if project.verbose:
    print('TABLE SCHEMA:', project_id, dataset_id, table_id)
  return API_BigQuery(auth).tables().get(
      projectId=project_id, datasetId=dataset_id,
      tableId=table_id).execute()['schema'].get('fields', [])


def table_to_type(auth, project_id, dataset_id, table_id):
  if project.verbose:
    print('TABLE TYPE:', project_id, dataset_id, table_id)
  return API_BigQuery(auth).tables().get(
      projectId=project_id, datasetId=dataset_id,
      tableId=table_id).execute()['type']


def query_to_rows(auth,
                  project_id,
                  dataset_id,
                  query,
                  row_max=None,
                  legacy=True,
                  as_object=False):
  if project.verbose:
    print('BIGQUERY QUERY:', project_id, dataset_id)

  # Create the query
  body = {
      'kind': 'bigquery#queryRequest',
      'query': query,
      'timeoutMs': 10000,
      'dryRun': False,
      'useQueryCache': True,
      'useLegacySql': legacy
  }

  if row_max:
    body['maxResults'] = row_max

  if dataset_id:
    body['defaultDataset'] = {'projectId': project_id, 'datasetId': dataset_id}

  # wait for query to complete

  response = API_BigQuery(auth).jobs().query(
      projectId=project_id, body=body).execute()
  while not response['jobComplete']:
    sleep(5)
    response = API_BigQuery(auth).jobs().getQueryResults(
        projectId=project_id,
        jobId=response['jobReference']['jobId']).execute(iterate=False)

  # fetch query results
  schema = response.get('schema', {}).get('fields', None)

  row_count = 0
  while 'rows' in response:

    for row in response['rows']:
      yield row_to_json(row, schema, as_object)
      row_count += 1

    if 'PageToken' in response:
      response = API_BigQuery(auth).jobs().getQueryResults(
          projectId=project_id,
          jobId=response['jobReference']['jobId'],
          pageToken=response['PageToken']).execute(iterate=False)
    elif row_count < int(response['totalRows']):
      response = API_BigQuery(auth).jobs().getQueryResults(
          projectId=project_id,
          jobId=response['jobReference']['jobId'],
          startIndex=row_count).execute(iterate=False)
    else:
      break


def make_schema(header):
  return [{
      'name': name,
      'type': 'STRING',
      'mode': 'NULLABLE'
  } for name in row_header_sanitize(header)]


def get_schema(rows, header=True, infer_type=True):
  """CAUTION: Memory suck.

  This function sabotages iteration by iterating thorough the new object and
  returning a new iterator RECOMMEND: Define the schema yourself, it will
  also ensure data integrity downstream.
  """

  schema = []
  row_buffer = []

  # everything else defaults to STRING
  type_to_bq = {
      int: 'INTEGER',
      bool: 'BOOLEAN',
      float: 'FLOAT'
  } if infer_type else {}  # empty lookup defaults to STRING below

  # first non null value determines type
  non_null_column = set()

  first = True
  ct_columns = 0

  for row in rows:

    # buffer the iterator to be returned with schema
    row += [None] * (ct_columns - len(row))
    row_buffer.append(row)

    # define schema field names and set defaults ( if no header enumerate fields )
    if first:
      ct_columns = len(row)
      for index, value in enumerate(row_header_sanitize(row)):
        schema.append({
            'name': value if header else 'Field_%d' % index,
            'type': 'STRING'
        })

    # then determine type of each column
    if not first and header:
      for index, value in enumerate(row):
        # if null, set only mode
        if value is None or value == '':
          schema[index]['mode'] = 'NULLABLE'
        else:
          column_type = type_to_bq.get(type(value), 'STRING')
          # if type is set, check to make sure its consistent
          if index in non_null_column:
            # change type only if its inconsistent
            if column_type != schema[index]['type']:
              # mixed integers and floats default to floats
              if column_type in (
                  'INTEGER', 'FLOAT') and schema[index]['type'] in ('INTEGER',
                                                                    'FLOAT'):
                schema[index]['type'] = 'FLOAT'
              # any strings are always strings
              else:
                schema[index]['type'] = 'STRING'
          # if first non null value, then just set type
          else:
            schema[index]['type'] = column_type
            non_null_column.add(index)

    # no longer first row
    first = False

  return row_buffer, schema


def drop_table(auth, project_id, dataset_id, table_id, billing_project_id=None):
  if not billing_project_id:
    billing_project_id = project_id

  query = ('DROP TABLE `' + project_id + '.' + dataset_id + '.' + table_id +
           '` ')

  body = {
      'kind': 'bigquery#queryRequest',
      'query': query,
      'defaultDataset': {
          'datasetId': dataset_id,
      },
      'useLegacySql': False,
  }

  job_wait(
      auth,
      API_BigQuery(auth).jobs().query(projectId=billing_project_id,
                                      body=body).execute())


def _get_max_date_from_table(auth,
                             project_id,
                             dataset_id,
                             table_id,
                             billing_project_id=None):
  if not billing_project_id:
    billing_project_id = project_id

  query = ('SELECT MAX(Report_Day) FROM `' + project_id + '.' + dataset_id +
           '.' + table_id + '` ')

  body = {
      'kind': 'bigquery#queryRequest',
      'query': query,
      'defaultDataset': {
          'datasetId': dataset_id,
      },
      'useLegacySql': False,
  }

  job = API_BigQuery(auth).jobs().query(
      projectId=billing_project_id, body=body).execute()
  return job['rows'][0]['f'][0]['v']


def _get_min_date_from_table(auth,
                             project_id,
                             dataset_id,
                             table_id,
                             billing_project_id=None):
  if not billing_project_id:
    billing_project_id = project_id

  query = ('SELECT MIN(Report_Day) FROM `' + project_id + '.' + dataset_id +
           '.' + table_id + '` ')

  body = {
      'kind': 'bigquery#queryRequest',
      'query': query,
      'defaultDataset': {
          'datasetId': dataset_id,
      },
      'useLegacySql': False,
  }

  job = API_BigQuery(auth).jobs().query(
      projectId=billing_project_id, body=body).execute()

  return job['rows'][0]['f'][0]['v']


#start and end date must be in format YYYY-MM-DD
def _clear_data_in_date_range_from_table(auth,
                                         project_id,
                                         dataset_id,
                                         table_id,
                                         start_date,
                                         end_date,
                                         billing_project_id=None):

  if not billing_project_id:
    billing_project_id = project_id

  query = ('DELETE FROM `' + project_id + '.' + dataset_id + '.' + table_id +
           '` ' + 'WHERE Report_Day >= "' + start_date + '"' +
           'AND Report_Day <= "' + end_date + '"')

  body = {
      'kind': 'bigquery#queryRequest',
      'query': query,
      'defaultDataset': {
          'datasetId': dataset_id,
      },
      'useLegacySql': False,
  }

  job_wait(
      auth,
      API_BigQuery(auth).jobs().query(projectId=billing_project_id,
                                      body=body).execute())



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import pprint
from time import sleep
from io import StringIO
from types import GeneratorType
from datetime import date, timedelta

from starthinker.config import BUFFER_SCALE
from starthinker.util import flag_last
from starthinker.util.data import get_rows
from starthinker.util.project import project
from starthinker.util.google_api import API_DCM
from starthinker.util.storage import media_download
from starthinker.util.csv import column_header_sanitize, csv_to_rows
from starthinker.util.dcm.schema.Lookup import DCM_Field_Lookup

DCM_CHUNK_SIZE = int(
    200 * 1024000 *
    BUFFER_SCALE)  # 200MB minimum recommended by docs * scale in config.py
DCM_CONVERSION_SIZE = 1000


def get_profile_for_api(auth, account_id):
  """Return a DCM profile ID for the currently supplied credentials.

  Bulletproofing:
  https://developers.google.com/doubleclick-advertisers/v3.2/userProfiles/get

  Handles cases of superuser, otherwise chooses the first matched profile.
  Allows DCM jobs to only specify account ID, which makes jobs more portable
  between accounts.

  Args:
    * auth: (string) Either user or service.
    * account_id: (int) Account number for which report is retrieved.

  Returns:
    * Is Superuser ( bool ): True if superuser account
    * Profile ID ( int ): profile id to be used to make API call

  Raises:
    * If current credentials do not have a profile for this account.

  """

  profile_admin = None
  profile_network = None

  for p in API_DCM(auth, iterate=True).userProfiles().list().execute():
    p_id = int(p['profileId'])
    a_id = int(p['accountId'])

    # take the first profile for admin
    if a_id == 2515 and 'subAccountId' not in p:
      profile_admin = p_id
      break

    # try to find a network profile if exists
    if a_id == account_id:
      profile_network = p_id
      break

  if profile_admin:
    return True, profile_admin
  elif profile_network:
    return False, profile_network
  else:
    raise Exception('Add your user profile to DCM account %s.' % account_id)


def get_account_name(auth, account):
  """ Return the name of a DCM account given the account ID.

  Args:
    * auth: (string) Either user or service.
    * account: (string) [account:advertiser@profile] token.

  Returns:
    * Profile ID.

  Raises:
    * If current credentials do not have a profile for this account.

  """

  account_id, advertiser_ids = parse_account(auth, account)
  is_superuser, profile_id = get_profile_for_api(auth, account_id)
  response = API_DCM(
      auth, internal=is_superuser).accounts().get(
          id=account_id, profileId=profile_id).execute()
  return response['name']


def parse_account(auth, account):
  """ Breaks a [account:advertiser@profile] string into parts if supplied.

  This function was created to accomodate supplying advertiser and profile
  information
  as a single token.  It needs to be refactored as this approach is messy.

  Possible variants include:
    * [account:advertiser@profile]
    * [account:advertiser]
    * [account@profile]

  Args:
    * auth: (string) Either user or service.
    * account: (string) A string represeting [account:advertiser@profile]

  Returns:
    * ( network_id, advertiser_ids, profile_id) after parsing the account token.

  """

  network_id = account
  advertiser_ids = None
  profile_id = None

  # if exists, get profile from end
  try:
    network_id, profile_id = network_id.split('@', 1)
  except:
    profile_id = None

  # if exists, get avertiser from end
  try:
    network_id, advertiser_ids = network_id.split(':', 1)
  except:
    pass

  # if network or advertiser, convert to integer
  if network_id is not None:
    network_id = int(network_id)
  if advertiser_ids is not None:
    advertiser_ids = [
        int(advertiser_id.strip())
        for advertiser_id in advertiser_ids.split(',')
    ]

  return network_id, advertiser_ids


def report_get(auth, account, report_id=None, name=None):
  """ Returns the DCM JSON definition of a report based on name or ID.

  Bulletproofing:
  https://developers.google.com/doubleclick-advertisers/v3.2/reports/get

  Args:
    * auth: (string) Either user or service.
    * account: (string) [account:advertiser@profile] token.
    * report_id: (int) ID of DCm report to fetch ( either or name ).
    * name: (string) Name of report to fetch ( either or report_id ).

  Returns:
    * JSON definition of report.

  """

  report = None

  account_id, advertiser_ids = parse_account(auth, account)
  is_superuser, profile_id = get_profile_for_api(auth, account_id)
  kwargs = {
      'profileId': profile_id,
      'accountId': account_id
  } if is_superuser else {
      'profileId': profile_id
  }

  if name:
    for r in API_DCM(
        auth, internal=is_superuser).reports().list(**kwargs).execute():
      if r['name'] == name:
        report = r
        break

  elif report_id:
    kwargs['reportId'] = report_id
    report = API_DCM(
        auth, internal=is_superuser).reports().get(**kwargs).execute()

  return report


def report_delete(auth, account, report_id=None, name=None):
  """ Deletes a DCM report based on name or ID.

  Bulletproofing:
  https://developers.google.com/doubleclick-advertisers/v3.2/reports/delete

  Args:
    * auth: (string) Either user or service.
    * account: (string) [account:advertiser@profile] token.
    * report_id: (int) ID of DCm report to fetch ( either or name ).
    * name: (string) Name of report to fetch ( either or report_id ).

  Returns:
    * None

  """

  report = report_get(auth, account, report_id, name)
  if report:
    account_id, advertiser_ids = parse_account(auth, account)
    is_superuser, profile_id = get_profile_for_api(auth, account_id)
    kwargs = {
        'profileId': profile_id,
        'accountId': account_id
    } if is_superuser else {
        'profileId': profile_id
    }
    kwargs['reportId'] = report['id']
    API_DCM(auth, internal=is_superuser).reports().delete(**kwargs).execute()
  else:
    if project.verbose:
      print('DCM DELETE: No Report')


def report_filter(auth, body, filters):
  """ Adds filters to a report body

  Filters cannot be easily added to the reports without templateing, this allows
  filters to be passed as lists.
  Values are specified using get_rows(...) helper, see
  starthinker/util/data/__init__.py.
  To specify a filter, use the official filter name and a list of values.

  For exmaple:

  ```
  filters = {
    "accountId": {
      "values": 789
    },
    "dfa:advertiser": {
      "values":[1234, 5678, 91011]
    }
  }
  ```

  Args:
    * auth: (string) Either user or service.
    * body: (json) the report body ( with or without filters )
    * filters: (json) a dictionary of filters to apply ( see above examples )

  Returns:
    * body: ( json ) modified report body
  """

  new_body = body.copy()

  for f, d in filters.items():
    for v in get_rows(auth, d):

      # accounts are specified in a unique part of the report json
      if f == 'accountId':
        new_body['accountId'] = v

      # activities are specified in a unique part of the report json
      elif f == 'dfa:activity':
        new_body['reachCriteria']['activities'].setdefault(
            'filters', []).append({
                'kind': 'dfareporting#dimensionValue',
                'dimensionName': f,
                'id': v
            })

      # all other filters go in the same place
      else:
        new_body.setdefault('criteria',
                            {}).setdefault('dimensionFilters', []).append({
                                'kind': 'dfareporting#dimensionValue',
                                'dimensionName': f,
                                'id': v,
                                'matchType': 'EXACT'
                            })

  return new_body


def report_build(auth, account, body):
  """ Creates a DCM report given a JSON definition.

  Bulletproofing:
  https://developers.google.com/doubleclick-advertisers/v3.2/reports/insert

  The body JSON provided will have the following fields overriden:
    * accountId - supplied as a parameter in account token.
    * ownerProfileId - determined from the current credentials.
    * advertiser_ids - supplied as a parameter in account token.

  Args:
    * auth: (string) Either user or service.
    * account: (string) [account:advertiser@profile] token.
    * body: (json) As defined in:
      https://developers.google.com/doubleclick-advertisers/v3.2/reports#resource

  Returns:
    * JSON definition of report created or existing.

  """

  report = report_get(auth, account, name=body['name'])

  if report is None:
    account_id, advertiser_ids = parse_account(auth, account)
    is_superuser, profile_id = get_profile_for_api(auth, account_id)

    # add the account id to the body
    body['accountId'] = account_id
    body['ownerProfileId'] = profile_id

    # add advertisers to the body, ignore for floodlight reports
    if advertiser_ids and 'criteria' in body:
      body['criteria']['dimensionFilters'] = body.get('criteria', {}).get(
          'dimensionFilters', []) + [{
              'kind': 'dfareporting#dimensionValue',
              'dimensionName': 'dfa:advertiser',
              'id': advertiser_id,
              'matchType': 'EXACT'
          } for advertiser_id in advertiser_ids]

    # add default daily schedule if it does not exist ( convenience )
    if 'schedule' not in body:
      body['schedule'] = {'active': True, 'repeats': 'DAILY', 'every': 1}

    # add default start and end if it does not exist ( convenience )
    if 'startDate' not in body['schedule']:
      body['schedule']['startDate'] = str(date.today())

    # add default start and end if it does not exist ( convenience )
    if 'expirationDate' not in body['schedule']:
      body['schedule']['expirationDate'] = str(
          (date.today() + timedelta(days=365)))

    #pprint.PrettyPrinter().pprint(body)

    # create the report
    kwargs = {
        'profileId': profile_id,
        'accountId': account_id
    } if is_superuser else {
        'profileId': profile_id
    }
    kwargs['body'] = body
    report = API_DCM(
        auth, internal=is_superuser).reports().insert(**kwargs).execute()

    # run the report
    kwargs = {
        'profileId': profile_id,
        'accountId': account_id
    } if is_superuser else {
        'profileId': profile_id
    }
    kwargs['reportId'] = report['id']
    API_DCM(auth, internal=is_superuser).reports().run(**kwargs).execute()

  else:
    if project.verbose:
      print('DCM Report Exists:', body['name'])

  return report


def report_fetch(auth, account, report_id=None, name=None, timeout=60):
  """ Retrieves most recent DCM file JSON by name or ID, if in progress, waits for it to complete.

  Bulletproofing:
  https://developers.google.com/doubleclick-advertisers/v3.2/files/get

  Timeout is in minutes ( retries will happen at 1 minute interval, default
  total time is 60 minutes )

  Args:
    * auth: (string) Either user or service.
    * account: (string) [account:advertiser@profile] token.
    * report_id: (int) ID of DCm report to fetch ( either or name ).
    * name: (string) Name of report to fetch ( either or report_id ).
    * timeout: (int) Minutes to wait for in progress report before giving up.

  Returns:
    * Report JSON if report exists and is ready.
    * True if report is in progress but not ready.
    * False if report does not exist.

  """

  if project.verbose:
    print('DCM REPORT FILE', report_id or name)

  if report_id is None:
    report = report_get(auth, account, name=name)
    if report is None:
      raise Exception('Report does not exist:', name)
    else:
      report_id = report['id']

  running = False

  # zero means run once
  while timeout >= 0:

    # loop all files recent to oldest looking for valid one
    for file_json in report_files(auth, account, report_id):
      #pprint.PrettyPrinter().pprint(file)

      # still running ( wait for timeout )
      if file_json['status'] == 'PROCESSING':
        if project.verbose:
          print('REPORT PROCESSING WILL WAIT')
        running = True
        if timeout > 0:
          break  # go to outer loop wait

      # ready for download ( return file )
      elif file_json['status'] == 'REPORT_AVAILABLE':
        if project.verbose:
          print('REPORT DONE')
        return file_json

      # cancelled or failed ( go to next file in loop )

    # if no report running ( skip wait )
    if not running:
      break

    # sleep a minute
    if timeout > 0:
      if project.verbose:
        print('WAITING MINUTES', timeout)
      sleep(60)

    # advance timeout
    timeout -= 1

  # if here, no file is ready, return status
  if project.verbose:
    print('NO REPORT FILES')
  return running


def report_run(auth, account, report_id=None, name=None):
  """ Trigger a DCM report to run by name or ID.

  Will do nothing if report is currently in progress.

  Bulletproofing:
  https://developers.google.com/doubleclick-advertisers/v3.3/reports/run

  Args:
    * auth: (string) Either user or service.
    * account: (string) [account:advertiser@profile] token.
    * report_id: (int) ID of DCm report to fetch ( either or name ).
    * name: (string) Name of report to fetch ( either or report_id ).

  Returns:
    * True if report run is executed
    * False otherwise

  """

  account_id, advertiser_id = parse_account(auth, account)
  is_superuser, profile_id = get_profile_for_api(auth, account_id)
  kwargs = {
      'profileId': profile_id,
      'accountId': account_id
  } if is_superuser else {
      'profileId': profile_id
  }

  if project.verbose:
    print('DCM REPORT RUN INIT', report_id or name)
  if report_id is None:
    report = report_get(auth, account, name=name)
    if report is None:
      raise Exception('Report does not exist:', name)
    else:
      report_id = report['id']

  kwargs = {
      'profileId': profile_id,
      'accountId': account_id
  } if is_superuser else {
      'profileId': profile_id
  }
  kwargs['reportId'] = report_id

  files = report_files(auth, account, report_id)
  latest_file_json = next(files, None)
  if latest_file_json == None or latest_file_json['status'] != 'PROCESSING':
    # run report if previously never run or currently not running
    if project.verbose:
      print('RUNNING REPORT', report_id or name)
    API_DCM(auth, internal=is_superuser).reports().run(**kwargs).execute()
    return True
  if project.verbose:
    print('REPORT RUN SKIPPED', report_id or name)
  return False


def report_file(auth,
                account,
                report_id=None,
                name=None,
                timeout=60,
                chunksize=DCM_CHUNK_SIZE):
  """ Retrieves most recent DCM file by name or ID, if in progress, waits for it to complete.

  Bulletproofing:
  https://developers.google.com/doubleclick-advertisers/v3.2/files/get

  Timeout is in minutes ( retries will happen at 1 minute interval, default
  total time is 60 minutes )
  If chunksize is set to 0 then the whole file is downloaded at once.

  Args:
    * auth: (string) Either user or service.
    * account: (string) [account:advertiser@profile] token.
    * report_id: (int) ID of DCm report to fetch ( either or name ).
    * name: (string) Name of report to fetch ( either or report_id ).
    * timeout: (int) Minutes to wait for in progress report before giving up.
    * chunksize: (int) number of bytes to download at a time, for memory
      constrained systems.

  Returns:
    * (filename, iterator) if file exists and is ready to download in chunks.
    * (filename, file) if file exists and chunking is off.
    * ('report_running.csv', None) if report is in progress.
    * (None, None) if file does not exist.

  """

  account_id, advertiser_id = parse_account(auth, account)
  file_json = report_fetch(auth, account, report_id, name, timeout)

  if file_json == False:
    return None, None
  elif file_json == True:
    return 'report_running.csv', None
  else:
    filename = '%s_%s.csv' % (file_json['fileName'],
                              file_json['dateRange']['endDate'].replace(
                                  '-', ''))

    # streaming
    if chunksize:
      return filename, media_download(
          API_DCM(auth).files().get_media(
              reportId=file_json['reportId'],
              fileId=file_json['id']).execute(False), chunksize, 'utf-8')

    # single object
    else:
      return filename, StringIO(
          API_DCM(auth).files().get_media(
              reportId=file_json['reportId'],
              fileId=file_json['id']).execute().decode('utf-8'))


def report_list(auth, account):
  """ Lists all the DCM report configurations for an account given the current credentials.

  Bulletproofing:
  https://developers.google.com/doubleclick-advertisers/v3.2/reports/list

  Args:
    * auth: (string) Either user or service.
    * account: (string) [account:advertiser@profile] token.

  Returns:
    * Iterator of JSONs.

  """

  account_id, advertiser_id = parse_account(auth, account)
  is_superuser, profile_id = get_profile_for_api(auth, account_id)
  kwargs = {
      'profileId': profile_id,
      'accountId': account_id
  } if is_superuser else {
      'profileId': profile_id
  }
  for report in API_DCM(
      auth, internal=is_superuser).reports().list(**kwargs).execute():
    yield report


def report_files(auth, account, report_id):
  """ Lists all the files available for a given DCM report configuration.

  Bulletproofing:
  https://developers.google.com/doubleclick-advertisers/v3.2/files/list

  Args:
    * auth: (string) Either user or service.
    * account: (string) [account:advertiser@profile] token.
    * report_id: (int) DCM report identifier.

  Returns:
    * Iterator of JSONs.

  """

  account_id, advertiser_id = parse_account(auth, account)
  is_superuser, profile_id = get_profile_for_api(auth, account_id)
  kwargs = {
      'profileId': profile_id,
      'accountId': account_id
  } if is_superuser else {
      'profileId': profile_id
  }
  kwargs['reportId'] = report_id
  for report_file in API_DCM(
      auth, internal=is_superuser).reports().files().list(**kwargs).execute():
    yield report_file


def report_to_rows(report):
  """ Helper to convert DCM files into iterator of rows, memory efficient.

  Usage example:

  ```
  filename, report = report_file(...)
  rows = report_to_rows(report)
  ```

  Args:
    * report: (iterator or file) Either an iterator or file that will be
      converted to rows.

  Returns:
    * Iterator of lists representing each row.

  """

  # if reading from stream
  if type(report) is GeneratorType:
    leftovers = ''
    for chunk in report:
      data, extra = chunk.rsplit('\n', 1)
      for row in csv_to_rows(leftovers + data):
        yield row
      leftovers = extra

  # if reading from buffer
  else:
    for row in csv_to_rows(report):
      yield row


def report_schema(headers):
  """ Helper to determine the schema of a given set of report headers.

  Using a match table generated from the DCM proto, each report header is
  matched
  to its type and a schema is assembled. If not found defaults to STRING.

  Usage example:

  ```
  filename, report = report_file(...)
  rows = report_to_rows(report)
  rows = report_clean(rows)
  schema = report_schema(next(rows))
  ```

  Args:
    * headers: (list) First row of a report.

  Returns:
    * JSON schema definition.

  """
  schema = []

  for header_name in headers:
    header_sanitized = column_header_sanitize(header_name)

    # first try exact match
    header_type = DCM_Field_Lookup.get(header_sanitized)

    # second try to match end for custom field names ( activity reports )
    if header_type is None:
      for field_name, field_type in DCM_Field_Lookup.items():
        if header_sanitized.endswith('_' + field_name):
          header_type = field_type
          break

    # finally default it to STRING
    if header_type is None:
      header_type = 'STRING'

    schema.append({
        'name': header_sanitized,
        'type': header_type,
        'mode': 'NULLABLE'
    })

  return schema


def report_clean(rows):
  """ Helper to fix DCM report issues for BigQuery and ensure schema compliance.

  Memory efficiently cleans each row by fixing:
  * Strips header and footer to preserve only data rows.
  * Changes 'Date' to 'Report_Day' to avoid using reserved name in BigQuery.
  * removes '-' as columns
  * Changes data format to match data studio if datastusio=True.

  Usage example:

  ```
  filename, report = report_file(...)
  rows = report_to_rows(report)
  rows = report_clean(rows)
  ```

  Args:
    * rows: (iterator) Rows to clean.

  Returns:
    * Iterator of cleaned rows.

  """

  if project.verbose:
    print('DCM REPORT CLEAN')

  first = True
  last = False

  # find start of report
  for row in rows:
    if row and row[0] == 'Report Fields':
      break

  # process the report
  for row in rows:
    # quit if empty report
    if 'No data returned by the reporting service.' in row:
      break

    # stop parsing if end of data
    if not row or row[0] == 'Grand Total:':
      break

    # find 'Date' column if it exists
    if first:
      try:
        date_column = row.index('Date')
        row[date_column] = 'Report_Day'
      except ValueError:
        pass
      row = [column_header_sanitize(cell) for cell in row]

    # remove not set columns ( which throw off schema on import types )
    row = ['' if cell.strip() in ('(not set)', '-') else cell for cell in row]

    # return the row
    yield row

    # not first row anymore
    first = False


def conversions_upload(auth,
                       account,
                       floodlight_activity_id,
                       conversion_type,
                       conversion_rows,
                       encryption_entity=None,
                       update=False):
  """ Uploads an offline conversion list to DCM.

  BulletProofing:
  https://developers.google.com/doubleclick-advertisers/guides/conversions_upload

  Handles errors and segmentation of conversion so list can be any size.

  Args:
    * auth: (string) Either user or service.
    * account: (string) [account:advertiser@profile] token.
    * floodlight_activity_id: (int) ID of DCM floodlight to upload conversions
      to.
    * converstion_type: (string) One of the following: encryptedUserId,
      encryptedUserIdCandidates, gclid, mobileDeviceId.
    * conversion_rows: (iterator) List of the following rows: Ordinal,
      timestampMicros, encryptedUserId | encryptedUserIdCandidates | gclid |
      mobileDeviceId.
    * encryption_entity: (object) See EncryptionInfo docs:
      https://developers.google.com/doubleclick-advertisers/v3.2/conversions/batchinsert#encryptionInfo
  """

  account_id, advertiser_id = parse_account(auth, account)
  is_superuser, profile_id = get_profile_for_api(auth, account_id)

  kwargs = {
      'profileId': profile_id,
      'accountId': account_id
  } if is_superuser else {
      'profileId': profile_id
  }
  kwargs['id'] = floodlight_activity_id
  response = API_DCM(
      auth,
      internal=is_superuser).floodlightActivities().get(**kwargs).execute()

  # upload in batch sizes of DCM_CONVERSION_SIZE
  row_count = 0
  row_buffer = []
  for is_last, row in flag_last(conversion_rows):
    row_buffer.append(row)

    if is_last or len(row_buffer) == DCM_CONVERSION_SIZE:

      if project.verbose:
        print('CONVERSION UPLOADING ROWS: %d - %d' %
              (row_count, row_count + len(row_buffer)))

      body = {
          'conversions': [{
              'floodlightActivityId':
                  floodlight_activity_id,
              'floodlightConfigurationId':
                  response['floodlightConfigurationId'],
              'ordinal':
                  row[0],
              'timestampMicros':
                  row[1],
              'quantity':
                  1,
              'value':
                  0.0,
              conversion_type:
                  row[2],
          } for row in row_buffer]
      }

      if encryption_entity:
        body['encryptionInfo'] = encryption_entity

      kwargs = {
          'profileId': profile_id,
          'accountId': account_id
      } if is_superuser else {
          'profileId': profile_id
      }
      kwargs['body'] = body

      if update:
        results = API_DCM(
            auth, internal=is_superuser).conversions().batchupdate(
                **kwargs).execute()
      else:
        results = API_DCM(
            auth, internal=is_superuser).conversions().batchinsert(
                **kwargs).execute()

      # stream back satus
      for status in results['status']:
        yield status

      # clear the buffer
      row_count += len(row_buffer)
      row_buffer = []


def id_to_timezone(reportGenerationTimeZoneId):
  return {
      1: 'America/New_York',
      2: 'Europe/London',
      3: 'Europe/Paris',
      4: 'Africa/Johannesburg',
      5: 'Asia/Jerusalem',
      6: 'Asia/Shanghai',
      7: 'Asia/Hong_Kong',
      8: 'Asia/Tokyo',
      9: 'Australia/Sydney',
      10: 'Asia/Dubai',
      11: 'America/Los_Angeles',
      12: 'Pacific/Auckland',
      13: 'America/Sao_Paulo',
  }.get(reportGenerationTimeZoneId)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import pandas
from datetime import date, timedelta

from starthinker.util.project import project
from starthinker.util.dcm import report_build, report_file, report_to_rows, report_clean, parse_account
from starthinker.util.sheets import sheets_tab_copy, sheets_read, sheets_url
from starthinker.util.csv import rows_to_type, rows_header_trim
from starthinker.util.email.template import EmailTemplate
from starthinker.util.email import send_email
"""Pulls floodlights from a sheet, checks if impressions have changed significantly and sends an alert email.

For example ( modify floodlight_monitor/test.json to include your account and
sheet ):

python floodlight_monitor/run.py floodlight_monitor/test.json -u [user
credentials path]

"""


def floodlight_report(floodlight_id):

  account_id, subaccount_id = parse_account(project.task['auth'],
                                            project.task['account'])

  name = 'Floodlight Monitor %s %s ( StarThinker )' % (account_id,
                                                       floodlight_id)

  if project.verbose:
    print('FLOODLIGHT MONITOR REPORT: ', name)

  # create report if it does not exists
  report = report_build(
      project.task['auth'], project.task['account'], {
          'kind':
              'dfareporting#report',
          'type':
              'FLOODLIGHT',
          'accountId':
              account_id,
          'name':
              name,
          'fileName':
              name.replace('( ', '').replace(' )', '').replace(' ', '_'),
          'format':
              'CSV',
          'delivery': {
              'emailOwner': False
          },
          'floodlightCriteria': {
              'dateRange': {
                  'kind': 'dfareporting#dateRange',
                  'relativeDateRange': 'LAST_7_DAYS'
              },
              'dimensions': [{
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:date'
              }, {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:floodlightConfigId'
              }, {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:activityGroupId'
              }, {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:activityGroup'
              }, {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:activityId'
              }, {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:activity'
              }],
              'floodlightConfigId': {
                  'dimensionName': 'dfa:floodlightConfigId',
                  'kind': 'dfareporting#dimensionValue',
                  'matchType': 'EXACT',
                  'value': floodlight_id
              },
              'metricNames': ['dfa:floodlightImpressions'],
              'reportProperties': {
                  'includeUnattributedCookieConversions': False,
                  'includeUnattributedIPConversions': False
              }
          },
          'schedule': {
              'active': True,
              'every': 1,
              'repeats': 'DAILY',
              'startDate': str(date.today()),
              'expirationDate': str((date.today() + timedelta(days=365))),
          },
      })
  return report['id']


def floodlight_rows(report_id):

  # fetch report file if it exists
  filename, report = report_file(project.task['auth'], project.task['account'],
                                 report_id, None, 10)

  # clean up rows
  rows = report_to_rows(report)
  rows = report_clean(rows)
  rows = rows_header_trim(rows)
  rows = rows_to_type(rows, column=6)

  return rows


def floodlight_outliers(df_in, col_name):
  q1 = df_in[col_name].quantile(0.25)
  q3 = df_in[col_name].quantile(0.75)
  iqr = q3 - q1
  fence_low = q1 - 1.5 * iqr
  fence_high = q3 + 1.5 * iqr

  df_in['status'] = 'NORMAL'
  df_in.loc[df_in[col_name] < fence_low, 'status'] = 'LOW'
  df_in.loc[df_in[col_name] > fence_high, 'status'] = 'HIGH'

  return df_in


def floodlight_analysis(rows):

  df = pandas.DataFrame.from_records(
      rows,
      columns=[
          'date', 'floodlightConfigId', 'activityGroupId', 'activityGroup',
          'activityId', 'activity', 'impressions'
      ])

  try:
    # return only lowest quartile outliers for the last day ( don't flag outliers from past days )
    last_day = pandas.to_datetime(df['date']).max().strftime('%Y-%m-%d')
    outliers = floodlight_outliers(
        df[[
            'date', 'floodlightConfigId', 'activityId', 'activity',
            'impressions'
        ]], 'impressions').values.tolist()
    outliers_today = [o for o in outliers if o[0] == last_day]
  except ValueError:
    print('No data available.')
    return None, None

  return last_day, outliers_today


def floodlight_email(day, alerts):

  for email, table in alerts.items():

    # build email template
    t = EmailTemplate()
    t.align('center')
    t.section(True)

    # when floodlight alerts exist
    issues = sum(1 for row in table if row[5] != 'NORMAL')

    if issues > 0:
      subject = '%d Floodlight Alerts For %s' % (issues, day)
    else:
      subject = 'All Floodlights Normal For %s' % day

    t.header(subject)
    t.paragraph(
        'The following floodlights are being monitored.  A status of LOW or HIGH inidcates impressions have changed significantly for the day.  A status of NORMAL means impressions are close to the average for the past 7 days.'
    )
    t.table([
        {
            'name': 'Date',
            'type': 'STRING'
        },
        {
            'name': 'Floodlight',
            'type': 'STRING'
        },
        {
            'name': 'Activity Id',
            'type': 'STRING'
        },
        {
            'name': 'Activity',
            'type': 'STRING'
        },
        {
            'name': 'Impressions',
            'type': 'INTEGER'
        },
        {
            'name': 'Status',
            'type': 'STRING'
        },
    ], table)

    t.paragraph(
        'Your monitored floodlights and recipients are listed in the sheet below.'
    )

    # either way link to the configuration sheet
    t.button(
        'Floodlight Monitoring Sheet',
        sheets_url(project.task['auth'], project.task['sheet']['sheet']),
        big=True)
    t.section(False)

    if project.verbose:
      print('FLOODLIGHT MONITOR EMAIL ALERTS', email, len(table))

    # send email template
    send_email(project.task['auth'], email, None, None, subject, t.get_text(),
               t.get_html())


@project.from_parameters
def floodlight_monitor():
  if project.verbose:
    print('FLOODLIGHT MONITOR')

  # make sure tab exists in sheet ( deprecated, use sheet task instead )
  if 'template' in project.task['sheet']:
    sheets_tab_copy(project.task['auth'],
                    project.task['sheet']['template']['sheet'],
                    project.task['sheet']['template']['tab'],
                    project.task['sheet']['sheet'],
                    project.task['sheet']['tab'])

  # read peers from sheet
  triggers = sheets_read(project.task['auth'], project.task['sheet']['sheet'],
                         project.task['sheet']['tab'],
                         project.task['sheet']['range'])
  # 0 - Floodlight Id
  # 1 - email
  # 2 - dcm report id ( added by this script )
  # 3 - status, added by the script ( LOW, NORMAL, HIGH )

  if project.verbose and len(triggers) == 0:
    print('FLOODLIGHT MONITOR: No floodlight ids specified in sheet.')

  alerts = {}
  day = None

  # create reports first in parallel
  for trigger in triggers:
    trigger.append(floodlight_report(trigger[0]))

  # download data from all reports
  for trigger in triggers:

    # get report rows for each floodlight
    rows = floodlight_rows(trigger[2])

    # calculate outliers
    last_day, rows = floodlight_analysis(rows)

    if last_day:
      # find last day report ran
      day = last_day if day is None else max(day, last_day)

      # group alerts by email
      alerts.setdefault(trigger[1], [])
      alerts[trigger[1]].extend(rows)

  if alerts:
    floodlight_email(day, alerts)


if __name__ == '__main__':
  floodlight_monitor()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of video formats."""

import sys

from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.feed import FieldMap


class VideoFormatDAO(BaseDAO):
  """Video format data access object.

  Inherits from BaseDAO and implements video format specific logic for creating
  and
  updating video format.
  """

  def __init__(self, auth, profile_id, is_admin):
    """Initializes VideoFormatDAO with profile id and authentication scheme."""
    super(VideoFormatDAO, self).__init__(auth, profile_id)

    self.profile_id = profile_id

    self._video_formats = None

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(VideoFormatDAO, self)._api(iterate).videoFormats()

  def get_video_formats(self):
    """Fetches video formats from CM.

    Returns:
      The lists of video formats from CM.
    """

    if not self._video_formats:
      self._video_formats = list(
          self._api(iterate=True).list(profileId=self.profile_id).execute())

    return self._video_formats

  def translate_transcode_config(self, transcode_configs):
    """Given a transcode config, returns the CM transcodes that match the config.

    Args:
      transcode_config: The transcode configuration feed item.

    Returns:
      All trancode objects from Campaign Manager that match the transcode
      configuration specified.
    """
    result = []
    REALLY_BIG_INT = 9223372036854775807

    try:
      for video_format in self.get_video_formats():
        for transcode_config in transcode_configs:
          min_width = int(transcode_config.get(FieldMap.TRANSCODE_MIN_WIDTH, 0))
          min_height = int(
              transcode_config.get(FieldMap.TRANSCODE_MIN_HEIGHT, 0))
          min_bitrate = int(
              transcode_config.get(FieldMap.TRANSCODE_MIN_BITRATE, 0))

          max_width = int(
              transcode_config.get(FieldMap.TRANSCODE_MAX_WIDTH,
                                   REALLY_BIG_INT))
          max_height = int(
              transcode_config.get(FieldMap.TRANSCODE_MAX_HEIGHT,
                                   REALLY_BIG_INT))
          max_bitrate = int(
              transcode_config.get(FieldMap.TRANSCODE_MAX_BITRATE,
                                   REALLY_BIG_INT))

          file_format = transcode_config.get(FieldMap.TRANSCODE_FORMAT, '')

          # There is one video format entry of id 15 in CM that represents the
          # Source File setting, it has no file type and no resolution, we are
          # using SOURCE_FILE as an artificial handle to allow this transcode
          # configuration to be selected from the sheet
          if file_format == 'SOURCE_FILE':
            if 15 not in result:
              result.append(15)
          elif min_width <= video_format['resolution']['width'] \
              and video_format['resolution']['width'] <= max_width \
              and min_height <= video_format['resolution']['height'] \
              and video_format['resolution']['height'] <= max_height \
              and min_bitrate <= video_format['targetBitRate'] \
              and video_format['targetBitRate'] <= max_bitrate \
              and video_format.get('fileType', '') == file_format:
            if video_format['id'] not in result:
              result.append(video_format['id'])
    except:
      raise Exception('Error determining file formats for transcode')

    return result
